"""Partly taken from bcbio-nextgen.
"""

import shutil
import os
from os.path import isfile, isdir, getsize, exists, basename, join, abspath, splitext, \
    islink, dirname, realpath, getmtime
import gzip
import tempfile
import contextlib
import itertools
import functools
import random
import collections
import fnmatch
import time
from functools import reduce

from ngs_utils.logger import info, err, warn, critical, debug

try:
    from concurrent import futures
except ImportError:
    try:
        import futures
    except ImportError:
        futures = None


@contextlib.contextmanager
def cpmap(cores=1):
    """Configurable parallel map context manager.

    Returns appropriate map compatible function based on configuration:
    - Local single core (the default)
    - Multiple local cores
    """
    if int(cores) == 1:
        yield itertools.imap
    else:
        if futures is None:
            raise ImportError("concurrent.futures not available")
        pool = futures.ProcessPoolExecutor(cores)
        yield pool.map
        pool.shutdown()

def map_wrap(f):
    """Wrap standard function to easily pass into 'map' processing.
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        return apply(f, *args, **kwargs)
    return wrapper


def safe_mkdir(dirpath, descriptive_name=''):
    if isdir(dirpath):
        return dirpath

    if not dirpath:
        critical(descriptive_name + ' path is empty.')

    if isfile(dirpath):
        critical(descriptive_name + ' ' + dirpath + ' is a file.')

    num_tries = 0
    max_tries = 10

    while not exists(dirpath):
        # we could get an error here if multiple processes are creating
        # the directory at the same time. Grr, concurrency.
        try:
            os.makedirs(dirpath)
        except OSError as e:
            if num_tries > max_tries:
                raise
            num_tries += 1
            time.sleep(2)
    return dirpath


def transform_to(ext):
    """
    Decorator to create an output filename from an output filename with
    the specified extension. Changes the extension, in_file is transformed
    to a new type.

    Takes functions like this to decorate:
    f(in_file, out_dir=None, out_file=None) or,
    f(in_file=in_file, out_dir=None, out_file=None)

    examples:
    @transform(".bam")
    f("the/input/path/file.sam") ->
        f("the/input/path/file.sam", out_file="the/input/path/file.bam")

    @transform(".bam")
    f("the/input/path/file.sam", out_dir="results") ->
        f("the/input/path/file.sam", out_file="results/file.bam")

    """

    def decor(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            out_file = kwargs.get("out_file", None)
            if not out_file:
                in_path = kwargs.get("in_file", args[0])
                out_dir = kwargs.get("out_dir", os.path.dirname(in_path))
                safe_mkdir(out_dir)
                out_name = replace_suffix(os.path.basename(in_path), ext)
                out_file = os.path.join(out_dir, out_name)
            kwargs["out_file"] = out_file
            if not file_exists(out_file):
                out_file = f(*args, **kwargs)
            return out_file
        return wrapper
    return decor


def filter_to(word):
    """
    Decorator to create an output filename from an input filename by
    adding a word onto the stem. in_file is filtered by the function
    and the results are written to out_file. You would want to use
    this over transform_to if you don't know the extension of the file
    going in. This also memoizes the output file.

    Takes functions like this to decorate:
    f(in_file, out_dir=None, out_file=None) or,
    f(in_file=in_file, out_dir=None, out_file=None)

    examples:
    @filter_to(".foo")
    f("the/input/path/file.sam") ->
        f("the/input/path/file.sam", out_file="the/input/path/file.foo.bam")

    @filter_to(".foo")
    f("the/input/path/file.sam", out_dir="results") ->
        f("the/input/path/file.sam", out_file="results/file.foo.bam")

    """

    def decor(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            out_file = kwargs.get("out_file", None)
            if not out_file:
                in_path = kwargs.get("in_file", args[0])
                out_dir = kwargs.get("out_dir", os.path.dirname(in_path))
                safe_mkdir(out_dir)
                out_name = append_stem(os.path.basename(in_path), word)
                out_file = os.path.join(out_dir, out_name)
            kwargs["out_file"] = out_file
            if not file_exists(out_file):
                out_file = f(*args, **kwargs)
            return out_file
        return wrapper
    return decor


def memoize_outfile(ext=None, stem=None):
    """
    Memoization decorator.

    See docstring for transform_to and filter_to for details.
    """
    if ext:
        return transform_to(ext)
    if stem:
        return filter_to(stem)


@contextlib.contextmanager
def chdir(new_dir):
    """Context manager to temporarily change to a new directory.

    http://lucentbeing.com/blog/context-managers-and-the-with-statement-in-python/
    """
    cur_dir = os.getcwd()
    safe_mkdir(new_dir)
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(cur_dir)

def file_uptodate(fname, cmp_fname):
    """Check if a file exists, is non-empty and is more recent than cmp_fname.
    """
    try:
        return (file_exists(fname) and file_exists(cmp_fname) and
                getmtime(fname) >= getmtime(cmp_fname))
    except OSError:
        return False

def create_dirs(config, names=None):
    if names is None:
        names = config["dir"].keys()
    for dname in names:
        d = config["dir"][dname]
        safe_mkdir(d)

def save_diskspace(fname, reason, config):
    """Overwrite a file in place with a short message to save disk.

    This keeps files as a sanity check on processes working, but saves
    disk by replacing them with a short message.
    """
    if config["algorithm"].get("save_diskspace", False):
        with open(fname, "w") as out_handle:
            out_handle.write("File removed to save disk space: %s" % reason)

def add_full_path(dirname, basedir=None):
    if basedir is None:
        basedir = os.getcwd()
    if not dirname.startswith("/"):
        dirname = os.path.join(basedir, dirname)
    return dirname

def symlink_plus(orig, new):
    """Create relative symlinks and handle associated biological index files.
    """
    for ext in ["", ".idx", ".gbi", ".tbi", ".bai"]:
        if os.path.exists(orig + ext) and not os.path.lexists(new + ext):
            with chdir(os.path.dirname(new)):
                os.symlink(os.path.relpath(orig + ext), os.path.basename(new + ext))
    orig_noext = splitext_plus(orig)[0]
    new_noext = splitext_plus(new)[0]
    for sub_ext in [".bai"]:
        if os.path.exists(orig_noext + sub_ext) and not os.path.lexists(new_noext + sub_ext):
            with chdir(os.path.dirname(new_noext)):
                os.symlink(os.path.relpath(orig_noext + sub_ext), os.path.basename(new_noext + sub_ext))

def open_gzipsafe(f, mode='r'):
    # mode_t = mode.replace('b', '')
    # mode_b = mode if 'b' in mode else mode + 'b'
    if f.endswith('.gz') or f.endswith('.gzip') or f.endswith('.gz.tx') or f.endswith('.gzip.tx'):
        try:
            h = gzip.open(f, mode=mode)
        except IOError as e:
            err('Error opening gzip ' + f + ': ' + str(e) + ', opening as plain text')
            return open(f, mode=mode)
        else:
            if 'w' in mode:
                return h
            else:
                try:
                    h.read(1)
                except IOError as e:
                    err('Error opening gzip ' + f + ': ' + str(e) + ', opening as plain text')
                    h.close()
                    return open(f, mode=mode)
                else:
                    h.close()
                    h = gzip.open(f, mode=mode)
                    return h
    else:
        return open(f, mode=mode)


def append_stem(to_transform, word):
    """
    renames a filename or list of filenames with 'word' appended to the stem
    of each one:
    example: append_stem("/path/to/test.sam", "_filtered") ->
    "/path/to/test_filtered.sam"

    """
    if is_sequence(to_transform):
        return [append_stem(f, word) for f in to_transform]
    elif is_string(to_transform):
        (base, ext) = splitext_plus(to_transform)
        return "".join([base, word, ext])
    else:
        raise ValueError("append_stem takes a single filename as a string or "
                         "a list of filenames to transform.")


def replace_suffix(to_transform, suffix):
    """
    replaces the suffix on a filename or list of filenames
    example: replace_suffix("/path/to/test.sam", ".bam") ->
    "/path/to/test.bam"

    """
    if is_sequence(to_transform):
        transformed = []
        for f in to_transform:
            (base, _) = os.path.splitext(f)
            transformed.append(base + suffix)
        return transformed
    elif is_string(to_transform):
        (base, _) = os.path.splitext(to_transform)
        return base + suffix
    else:
        raise ValueError("replace_suffix takes a single filename as a string or "
                         "a list of filenames to transform.")

# ## Functional programming

def partition_all(n, iterable):
    """Partition a list into equally sized pieces, including last smaller parts
    http://stackoverflow.com/questions/5129102/python-equivalent-to-clojures-partition-all
    """
    it = iter(iterable)
    while True:
        chunk = list(itertools.islice(it, n))
        if not chunk:
            break
        yield chunk

def partition(pred, iterable):
    'Use a predicate to partition entries into false entries and true entries'
    # partition(is_odd, range(10)) --> 0 2 4 6 8   and  1 3 5 7 9
    t1, t2 = itertools.tee(iterable)
    try:
        return itertools.ifilterfalse(pred, t1), itertools.ifilter(pred, t2)
    except:
        return itertools.filterfalse(pred, t1), filter(pred, t2)

# ## Dealing with configuration files

def merge_config_files(fnames):
    """Merge configuration files, preferring definitions in latter files.
    """
    import yaml
    def _load_yaml(fname):
        with open(fname) as in_handle:
            config = yaml.load(in_handle)
        return config
    out = _load_yaml(fnames[0])
    for fname in fnames[1:]:
        cur = _load_yaml(fname)
        for k, v in cur.items():
            if k in out and isinstance(out[k], dict):
                out[k].update(v)
            else:
                out[k] = v
    return out


def get_in(d, t, default=None):
    """
    look up if you can get a tuple of values from a nested dictionary,
    each item in the tuple a deeper layer

    example: get_in({1: {2: 3}}, (1, 2)) -> 3
    example: get_in({1: {2: 3}}, (2, 3)) -> {}
    """
    result = reduce(lambda d, t: d.get(t, {}), t, d)
    if not result:
        return default
    else:
        return result


def flatten(l):
    """
    flatten an irregular list of lists
    example: flatten([[[1, 2, 3], [4, 5]], 6]) -> [1, 2, 3, 4, 5, 6]
    lifted from: http://stackoverflow.com/questions/2158395/

    """
    for el in l:
        if isinstance(el, collections.Iterable) and not isinstance(el,
                                                                   str):
            for sub in flatten(el):
                yield sub
        else:
            yield el


def is_sequence(arg):
    """
    check if 'arg' is a sequence

    example: arg([]) -> True
    example: arg("lol") -> False

    """
    return (not hasattr(arg, "strip") and
            hasattr(arg, "__getitem__") or
            hasattr(arg, "__iter__"))


def is_pair(arg):
    """
    check if 'arg' is a two-item sequence

    """
    return is_sequence(arg) and len(arg) == 2

def is_string(arg):
    return isinstance(arg, str)


def locate(pattern, root=os.curdir):
    '''Locate all files matching supplied filename pattern in and below
    supplied root directory.'''
    for path, dirs, files in os.walk(os.path.abspath(root)):
        for filename in fnmatch.filter(files, pattern):
            yield os.path.join(path, filename)


def itersubclasses(cls, _seen=None):
    """
    snagged from:  http://code.activestate.com/recipes/576949/
    itersubclasses(cls)

    Generator over all subclasses of a given class, in depth first order.

    >>> list(itersubclasses(int)) == [bool]
    True
    >>> class A(object): pass
    >>> class B(A): pass
    >>> class C(A): pass
    >>> class D(B,C): pass
    >>> class E(D): pass
    >>>
    >>> for cls in itersubclasses(A):
    ...     print(cls.__name__)
    B
    D
    E
    C
    >>> # get ALL (new-style) classes currently defined
    >>> [cls.__name__ for cls in itersubclasses(object)] #doctest: +ELLIPSIS
    ['type', ...'tuple', ...]
    """

    if not isinstance(cls, type):
        raise TypeError('itersubclasses must be called with '
                        'new-style classes, not %.100r' % cls)
    if _seen is None:
        _seen = set()
    try:
        subs = cls.__subclasses__()
    except TypeError:  # fails only when cls is type
        subs = cls.__subclasses__(cls)
    for sub in subs:
        if sub not in _seen:
            _seen.add(sub)
            yield sub
            for sub in itersubclasses(sub, _seen):
                yield sub

def replace_directory(out_files, dest_dir):
    """
    change the output directory to dest_dir
    can take a string (single file) or a list of files
    """
    if is_sequence(out_files):
        filenames = map(os.path.basename, out_files)
        return [os.path.join(dest_dir, x) for x in filenames]
    elif is_string(out_files):
        return os.path.join(dest_dir, os.path.basename(out_files))
    else:
        raise ValueError("in_files must either be a sequence of filenames "
                         "or a string")

def which(program):
    """
    returns the path to an executable or None if it can't be found
    """
    def is_exe(_fpath):
        return os.path.isfile(_fpath) and os.access(_fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file
    return None

def reservoir_sample(stream, num_items, item_parser=lambda x: x):
    """
    samples num_items from the stream keeping each with equal probability
    """
    kept = []
    for index, item in enumerate(stream):
        if index < num_items:
            kept.append(item_parser(item))
        else:
            r = random.randint(0, index)
            if r < num_items:
                kept[r] = item_parser(item)
    return kept


def compose(f, g):
    return lambda x: f(g(x))

def dictapply(d, fn):
    """
    apply a function to all non-dict values in a dictionary
    """
    for k, v in d.items():
        if isinstance(v, dict):
            v = dictapply(v, fn)
        else:
            d[k] = fn(v)
    return d


def verify_module(name):
    try:
        __import__(name)
        return True
    except:
        return False


def adjust_path(path):
    if path is None: return None

    path = remove_quotes(path)
    if path is None: return None

    path = expanduser(path)
    if path is None: return None

    path = abspath(path)
    if path is None: return None

    return path


code_base_path = abspath(join(dirname(abspath(__file__)), os.path.pardir))


def adjust_system_path(path):
    if path is None: return None

    path = remove_quotes(path)
    if path is None: return None

    path = expanduser(path)
    if path is None: return None

    path = join(code_base_path, path)  # will only join if the tool_path is not absolute:
    if path is None: return None

    path = realpath(path)
    if path is None: return None

    path = abspath(path)
    if path is None: return None

    return path


# Expand paths beginning with '~' or '~user'.
# '~' means $HOME; '~user' means that user's home directory.
# If the path doesn't begin with '~', or if the user or $HOME is unknown,
# the path is returned unchanged (leaving error reporting to whatever
# function is called with the expanded path as argument).
# See also module 'glob' for expansion of *, ? and [...] in pathnames.
# (A function should also be defined to do full *sh-style environment
# variable expansion.)
def expanduser(path):
    """Expand ~ and ~user constructs.

    If user or $HOME is unknown, do nothing."""
    if path[:1] != '~':
        return path
    i, n = 1, len(path)
    while i < n and path[i] not in '/\\':
        i = i + 1

    if 'HOME' in os.environ:
        userhome = os.environ['HOME']
    elif 'USERPROFILE' in os.environ:
        userhome = os.environ['USERPROFILE']
    elif not 'HOMEPATH' in os.environ:
        return path
    else:
        try:
            drive = os.environ['HOMEDRIVE']
        except KeyError:
            drive = ''
        userhome = join(drive, os.environ['HOMEPATH'])

    if i != 1:  # ~user
        userhome = join(dirname(userhome), path[1:i])

    return userhome + path[i:]


def file_exists(fpath):
    """Check if a file exists and is non-empty.
    """
    return fpath and exists(adjust_path(fpath)) and getsize(adjust_path(fpath)) > 0


def _log(msg, silent, is_critical):
    if is_critical:
        critical(msg)
    if not silent:
        warn(msg)

def verify_obj_by_path(path, description='', silent=False, is_critical=False, verify_size=True):
    if path is None:
        msg = (description + ': i' if description else 'I') + 's not specified (None).'
        _log(msg, silent, is_critical)
        return path

    if not path:
        msg = (description + ': f' if description else 'N') + 'ame is empty.'
        _log(msg, silent, is_critical)
        return path

    path = adjust_path(path)
    if not exists(path):
        msg = (description + ': ' if description else '') + path + ' does not exist.'
        _log(msg, silent, is_critical)
        return None

    if isfile(path):
        return verify_file(path, description, silent, verify_size=verify_size)
    elif isdir(path):
        return verify_dir(path, description, silent)
    else:
        msg = (description + ': ' if description else '') + path + ' is not a file or a directory.'
        _log(msg, silent, is_critical)
        return None

def can_reuse(fpath, cmp_f, silent=False):
    do_reuse = os.environ.get('REUSE', '1')
    if do_reuse == '0':
        return False
    if not fpath or not isfile(fpath):
        return False
    elif verify_file(fpath, cmp_f=cmp_f, silent=True):
        if not silent:
            debug('Reusing ' + fpath)
        return True
    else:
        return False

def verify_file(fpath, description='', silent=False, is_critical=False, verify_size=True, cmp_f=None):
    if fpath is None:
        msg = (description + ': ' if description else ' ') + 'not specified.'
        _log(msg, silent, is_critical)
        return fpath

    if not fpath:
        msg = (description + ': f' if description else 'F') + 'ile name is empty.'
        _log(msg, silent, is_critical)
        return fpath

    fpath = adjust_path(fpath)
    if not exists(fpath):
        msg = (description + ': ' if description else '') + fpath + ' does not exist.'
        _log(msg, silent, is_critical)
        return None

    if not isfile(fpath):
        msg = (description + ': ' if description else '') + fpath + ' is not a file.'
        _log(msg, silent, is_critical)
        return None

    if verify_size and getsize(fpath) <= 0:
        msg = (description + ': ' if description else '') + fpath + ' is empty.'
        _log(msg, silent, is_critical)
        return None

    if cmp_f:
        if isinstance(cmp_f, str):
            cmp_f = [cmp_f]
        try:
            for cmp_f in cmp_f:
                if cmp_f and getmtime(fpath) < getmtime(cmp_f):
                    msg = (description + ': ' if description else '') + fpath + ' is older than ' + cmp_f
                    _log(msg, silent, is_critical)
                    return None
        except OSError as e:
            _log(str(e), silent, is_critical)

    return fpath

def verify_dir(dirpath, description='', silent=False, is_critical=False):
    if dirpath is None:
        msg = (description + ': i' if description else 'I') + 's not specified (None).'
        _log(msg, silent, is_critical)
        return dirpath

    if not dirpath:
        msg = (description + ': d' if description else 'D') + 'ir name is empty.'
        _log(msg, silent, is_critical)
        return None

    dirpath = adjust_path(dirpath)
    if not exists(dirpath):
        msg = (description + ': ' if description else '') + dirpath + ' does not exist.'
        _log(msg, silent, is_critical)
        return None

    if not isdir(dirpath):
        msg = (description + ': ' if description else '') + dirpath + ' is not a directory.'
        _log(msg, silent, is_critical)
        return None

    return dirpath


def num_lines(fpath):
    with open(fpath) as f:
        return sum(1 for l in f)


def make_tmpdir():
    # base_dir = cnf.tmp_base_dir or cnf.work_dir or os.getcwd()
    # if not verify_dir(base_dir, 'Base directory for temporary files'):
    #     sys.exit(1)
    #
    return tempfile.mkdtemp()


@contextlib.contextmanager
def tmpdir():
    dirpath = make_tmpdir()
    try:
        yield dirpath
    finally:
        try:
            shutil.rmtree(dirpath)
        except OSError:
            warn('Warning: cannot clean up temporary dir ' + dirpath)


@contextlib.contextmanager
def workdir(cnf):
    if cnf.work_dir:
        verify_dir(cnf.work_dir, is_critical=True)
        yield cnf.work_dir
    else:
        cnf.work_dir = make_tmpdir()
        yield cnf.work_dir
        try:
            shutil.rmtree(cnf.work_dir)
        except OSError:
            warn('Warning: cannot clean up temporary dir ' + cnf.work_dir)


def make_tmpfile(work_dir, *args, **kwargs):
    """ Returns tuple (file descriptor and file path)
    """
    return tempfile.mkstemp(work_dir, *args, **kwargs)


@contextlib.contextmanager
def tmpfile(work_dir, *args, **kwargs):
    tmp_file, fpath = make_tmpfile(work_dir, *args, **kwargs)
    try:
        yield fpath
    finally:
        try:
            os.remove(fpath)
        except OSError:
            pass


def splitext_plus(fname):
    """Split on file extensions, allowing for zipped extensions.
    """
    base, ext = splitext(fname)
    if ext in [".gz", ".bz2", ".zip"]:
        base, ext2 = splitext(base)
        ext = ext2 + ext
    return base, ext


def add_suffix(fname, suf):
    base, ext = splitext_plus(fname)
    return base + (('.' + suf) if suf else '') + ext


def intermediate_fname(work_dir, fpath, suf):
    suf_fpath = add_suffix(fpath, suf)
    if work_dir:
        return join(work_dir, basename(suf_fpath))
    else:
        return suf_fpath


def remove_quotes(s):
    if s and s[0] in ['"', "'"]:
        s = s[1:]
    if s and s[-1] in ['"', "'"]:
        s = s[:-1]
    return s


def convert_file(work_dir, input_fpath, convert_file_fn, suffix=None, output_fpath=None,
                 check_result=True, overwrite=False, reuse=True, ctx=None):
    assert output_fpath or suffix, str(output_fpath) + ' ' + str(suffix)
    output_fpath = output_fpath or intermediate_fname(work_dir, input_fpath, suf=suffix)
    if output_fpath.endswith('.gz'):
        debug('output_fpath is .gz, but writing to uncompressed.')
        output_fpath = splitext(output_fpath)[0]
    
    if not overwrite:
        if can_reuse(output_fpath, cmp_f=input_fpath):
            debug('Reusing ' + output_fpath)
            return output_fpath
        if can_reuse(output_fpath + '.gz', cmp_f=input_fpath):
            debug('Reusing ' + output_fpath + '.gz')
            return output_fpath
    
    if islink(output_fpath):
        os.unlink(output_fpath)

    debug('Writing to ' + output_fpath)
    with file_transaction(work_dir, output_fpath) as tx_fpath:
        with open_gzipsafe(input_fpath) as inp_f, open(tx_fpath, 'w') as out_f:
            if ctx:
                convert_file_fn(inp_f, out_f, ctx)
            else:
                convert_file_fn(inp_f, out_f)

    if suffix or output_fpath:
        debug('Saved to ' + output_fpath)

    verify_file(output_fpath, is_critical=check_result)
    return output_fpath


def iterate_file(work_dir, input_fpath, proc_line_fun,
                 suffix=None, output_fpath=None, check_result=True, reuse=False, **kwargs):
    def _proc_file(inp_f, out_f, ctx=None):
        max_bunch_size = 1000 * 1000
        written_lines = 0
        bunch = []

        for i, line in enumerate(inp_f):
            clean_line = line.replace('\n', '')
            if clean_line:
                if ctx:
                    new_l = proc_line_fun(clean_line, i, ctx)
                else:
                    new_l = proc_line_fun(clean_line, i)
                if new_l is not None:
                    bunch.append(new_l + '\n')
                    written_lines += 1
            else:
                bunch.append(line)
                written_lines += 1

            if len(bunch) >= max_bunch_size:
                out_f.writelines(bunch)
                debug('Written lines: ' + str(written_lines))
                bunch = []

        out_f.writelines(bunch)
        debug('Written lines: ' + str(written_lines))

    return convert_file(work_dir, input_fpath, _proc_file,
        suffix=suffix, output_fpath=output_fpath, check_result=check_result, reuse=reuse, **kwargs)


def dots_to_empty_cells(config, tsv_fpath):
    """Put dots instead of empty cells in order to view TSV with column -t
    """
    def proc_line(l, i):
        while '\t\t' in l:
            l = l.replace('\t\t', '\t.\t')
        return l
    return iterate_file(config, tsv_fpath, proc_line, suffix='dots')


def __remove_tmpdirs(fnames):
    if isinstance(fnames, str):
        fnames = [fnames]
    for x in fnames:
        xdir = os.path.dirname(os.path.abspath(x))
        if xdir and os.path.exists(xdir):
            shutil.rmtree(xdir, ignore_errors=True)


def __remove_files(fnames):
    if isinstance(fnames, str):
        fnames = [fnames]

    for x in fnames:
        if x and os.path.exists(x):
            if os.path.isfile(x):
                os.remove(x)
            elif os.path.isdir(x):
                shutil.rmtree(x, ignore_errors=True)


#################################################
######## Transaction ############################
@contextlib.contextmanager
def file_transaction(work_dir, *rollback_files):
    """Wrap file generation in a transaction, moving to output if finishes.
    """
    exts = {".vcf": ".idx", ".bam": ".bai", "vcf.gz": ".tbi"}
    safe_fpaths, orig_names = _flatten_plus_safe(work_dir, rollback_files)
    __remove_files(safe_fpaths)  # remove any half-finished transactions
    try:
        if len(safe_fpaths) == 1:
            yield safe_fpaths[0]
        else:
            yield tuple(safe_fpaths)
    except:  # failure -- delete any temporary files
        __remove_files(safe_fpaths)
        raise
    else:  # worked -- move the temporary files to permanent location
        for safe, orig in zip(safe_fpaths, orig_names):
            if exists(safe):
                shutil.move(safe, orig)
                for check_ext, check_idx in exts.items():
                    if safe.endswith(check_ext):
                        safe_idx = safe + check_idx
                        if exists(safe_idx):
                            shutil.move(safe_idx, orig + check_idx)


@contextlib.contextmanager
def tx_tmpdir(base_dir, rollback_dirpath):
    """Context manager to create and remove a transactional temporary directory.
    """
    # tmp_dir_base = join(base_dir, 'tx', str(uuid.uuid4()))
    # unique_attempts = 0
    # while os.path.exists(tmp_dir_base):
    #     if unique_attempts > 5:
    #         break
    #     tmp_dir_base = join(base_dir, 'tx', str(uuid.uuid4()))
    #     time.sleep(1)
    #     unique_attempts += 1

    # if base_dir is not None:
    #     tmp_dir_base = os.path.join(base_dir, "tx")
    # else:
    #     tmp_dir_base = os.path.join(os.getcwd(), "tx")
    if exists(rollback_dirpath):
        critical(rollback_dirpath + ' already exists')

    tmp_dir = tempfile.mkdtemp(dir=base_dir)
    safe_mkdir(tmp_dir)
    try:
        yield tmp_dir
    finally:
        if tmp_dir and exists(tmp_dir):
            os.rename(tmp_dir, rollback_dirpath)


def _flatten_plus_safe(tmp_dir, rollback_files):
    """Flatten names of files and create temporary file names.
    """
    tx_fpaths, orig_files = [], []
    for fnames in rollback_files:
        if isinstance(fnames, str):
            fnames = [fnames]
        for fname in fnames:
            tx_file = fname + '.tx'
            tx_fpath = join(tmp_dir, tx_file) if tmp_dir else tx_file
            tx_fpaths.append(tx_fpath)
            orig_files.append(fname)
    return tx_fpaths, orig_files


@contextlib.contextmanager
def chdir(new_dir):
    """Context manager to temporarily change to a new directory.
    http://lucentbeing.com/blog/context-managers-and-the-with-statement-in-python/
    """
    cur_dir = os.getcwd()
    safe_mkdir(new_dir)
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(cur_dir)


def safe_symlink_to(fpath, dst_dirpath):
    dst = join(dst_dirpath, basename(fpath))
    if not exists(dst):
        try:
            if os.lstat(dst):  # broken symlink
                os.remove(dst)
        except OSError:
            pass
        debug('Symlink ' + fpath + ' -> ' + dst)
        os.symlink(fpath, dst)
    return dst


def safe_symlink(src_path, dst_path):
    if not exists(dst_path):
        try:
            if os.lstat(dst_path):  # broken symlink
                os.remove(dst_path)
        except:
            pass
        os.symlink(src_path, dst_path)
    return dst_path


def is_gz(fpath, mode='rb'):
    try:
        h = gzip.open(fpath)
    except IOError as e:
        return False
    else:
        try:
            h.read(1)
        except IOError as e:
            h.close()
            return False
        else:
            h.close()
            return True




#!/usr/bin/env python
'''
Combines multiple tab-delimited files into one XLSX file
'''

import sys, os
import xlsxwriter

from .support import gzip_opener

def tab_combine(outfile, fnames ,delim='\t'):
    workbook = xlsxwriter.Workbook(outfile, {'strings_to_numbers': True})

    for fname in fnames:
        sys.stderr.write(fname.rstrip(".gz").rstrip(".txt"))
        sys.stderr.write("... ")
        worksheet = workbook.add_worksheet(fname.rstrip(".gz").rstrip(".txt")[:31])
        f = gzip_opener(fname).open()
        row = 0
        for line in f:
            cols = line.rstrip().split(delim)
            for col, val in enumerate(cols):
                worksheet.write(row, col, val)
            row += 1
        sys.stderr.write("done\n")
        f.close()

    workbook.close()

 
    
def usage(msg=""):
    if msg:
        print(msg)
    print(__doc__)
    print("""Usage: %s {opts} outfile.xlsx filename1.tab filename2...

Options:
    -f          Force overwriting the output file.
    -d delim    Use this (opposed to a tab) for the delimiter

""" % os.path.basename(sys.argv[0]))
    sys.exit(1)
    
def main(argv):
    outfile = None
    fnames = []

    delim = '\t'
    force = False
    last = None
    for arg in argv:
        if arg in ['-h','--help']:
            usage()
        elif last == '-d':
            delim = arg
            last = None
        elif arg in ['-d']:
            last = arg
        elif arg == '-f':
            force = True;
        elif not outfile:
            outfile = arg
        elif os.path.exists(arg):
            fnames.append(arg)

    if outfile and os.path.exists(outfile) and not force:
        usage("%s exists! If you want to overwrite this file, use the -f (force) option!" % outfile)

    if not outfile or not fnames:
        usage("Missing output file or input filename(s)!")

    tab_combine(outfile, fnames, delim)
    
if __name__ == '__main__':
    main(sys.argv[1:])




import hashlib
from os import environ
import socket
import re
from collections import OrderedDict
from os.path import join


class OrderedDefaultDict(OrderedDict):
    def __init__(self, *args, **kwargs):
        if not args:
            self.default_factory = None
        else:
            if not (args[0] is None or callable(args[0])):
                raise TypeError('first argument must be callable or None')
            self.default_factory = args[0]
            args = args[1:]
        super(OrderedDefaultDict, self).__init__(*args, **kwargs)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = default = self.default_factory()
        return default

    def __reduce__(self):  # optional, for pickle support
        args = (self.default_factory,) if self.default_factory else ()
        return self.__class__, args, None, None, self.items()


def _tryint(s):
    try:
        return int(s)
    except ValueError:
        return s


def _alphanum_key(s):
    """ Turn a string into a list of string and number chunks.
        "z23a" -> ["z", 23, "a"]
    """
    return [_tryint(c) for c in re.split('([0-9]+)', s)]


def human_sorted(l):
    """ Sort the given list in the way that humans expect.
    """
    l.sort(key=_alphanum_key)
    return l


def format_integer(name, value, unit=''):
    value = int(value)
    if value is not None:
        return '{name}: {value:,}{unit}'.format(**locals())
    else:
        return '{name}: -'.format(**locals())


def format_decimal(name, value, unit=''):
    if value is not None:
        return '{name}: {value:.2f}{unit}'.format(**locals())
    else:
        return '{name}: -'.format(**locals())


def mean(values):
    return float(sum(values)) / len(values) if len(values) > 0 else float('nan')


def median(values):
    values = sorted(values)

    if len(values) % 2 == 1:  # odd number of values
        return values[(len(values) - 1) // 2]
    else:  # even number of values - take the avg of central
        return (values[len(values) // 2] + values[len(values) // 2 - 1]) // 2


def get_numeric_value(string_value):
    """ parses string_value and returns only number-like part
    """
    num_chars = ['.', '+', '-']
    number = ''
    for c in string_value:
        if c.isdigit() or c in num_chars:
            number += c
    return number


hostname = socket.gethostname()

def is_us():
    return any(name in hostname for name in ['rask', 'chara', 'blue', 'green', 'espo',
                                             'orr', 'usbod', 'bn0', 'pedro', 'papi'])
def is_uk():
    return 'ukap' in hostname

def is_sweden():
    return 'seml' in hostname

def is_china():
    return 'cniclhpc' in hostname

def is_az():
    return is_us() or is_uk() or is_china() or is_sweden()

def is_cloud():
    return 'starcluster' in hostname

def is_cluster():
    return is_az() or is_cloud()

def is_travis():
    return 'TRAVIS' in environ

def is_chihua():
    return hostname == 'chihua'

def is_local():
    return 'Vlads-MBP' in hostname or 'Vladislavs-Air.lan' in hostname or 'local' in hostname or 'Home' in hostname or environ.get('PYTHONUNBUFFERED')


def md5(fpath):
    hash = hashlib.md5()
    with open(fpath, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash.update(chunk)
    return hash.hexdigest()


def gray(text):
    return '<span class="gray">' + text + '</span>'



import getpass
import os

from os import environ
import socket
import sys
from datetime import datetime
from time import sleep
import smtplib
from email.mime.text import MIMEText
import traceback
from subprocess import check_output

from ngs_utils.utils import is_cluster, is_local

log_fpath = None
project_name = None
project_fpath = None
proc_name = None
is_debug = False

smtp_host = None  # set up in source/config.py and system_info.yaml
my_address = 'Vlad.Saveliev@astrazeneca.com'


error_msgs = []
warning_msgs = []
critical_msgs = []


def init(is_debug_=None, log_fpath_=None):
    if is_debug_:
        global is_debug
        is_debug = is_debug_
    if log_fpath_:
        set_log_path(log_fpath_)
    info(check_output('hostname', shell=True).strip())
    with open(os.devnull, 'w') as devnull:
        username = check_output('finger $(whoami) | head -n1', shell=True, stderr=devnull).strip()
        if not username:
            username = getpass.getuser()
        info(username)
    info()
    info(' '.join(sys.argv))
    info()
    info('-' * 70)


past_msgs = []
def set_log_path(log_fpath_):
    assert log_fpath_
    global log_fpath, past_msgs
    log_fpath = log_fpath_
    for msg in past_msgs:
        _write_to_file(msg)
    past_msgs = []
    
    
def set_smtp_host(smtp_host_):
    if smtp_host_:
        global smtp_host
        smtp_host = smtp_host_
        debug('Set smtp host ' + smtp_host)


def timestamp():
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


def step_greetings(name):
    info()
    info('-' * 70)
    info(name)
    info('-' * 70)


def info(msg='', ending='\n', print_date=True, severity='info'):
    _log(sys.stdout, msg, ending, print_date, severity=severity)


def debug(msg='', ending='\n', print_date=True, severity='debug'):
    _log(sys.stdout, '[DEBUG] ' + msg, ending, print_date, severity=severity)


def warn(msg='', ending='\n', print_date=True, severity='warning'):
    _log(sys.stderr, msg, ending, print_date, severity=severity)


def silent_err(msg='', ending='\n', print_date=True, severity='silent_err'):
    warn(msg, ending, print_date, severity=severity)


def err(msg='', ending='\n', print_date=True, severity='error'):
    warn(msg, ending, print_date, severity=severity)
error = err


def critical(msg=''):
    if isinstance(msg, str):
        err(msg, severity='critical')
    else:
        if not msg:
            return
        for m in msg:
            err(m, severity='critical')
    raise CriticalError(msg)


def send_email(msg_other='', subj='', only_me=False, addr_by_username=None, addr=None):
    if not msg_other or not smtp_host:
        return
    debug('Emailing using smtp host ' + smtp_host)

    username = getpass.getuser()

    other_address = None
    if not only_me:
        if addr:
            other_address = addr
        elif addr_by_username and username in addr_by_username:
            other_address = addr_by_username[username]
    if other_address == my_address:
        other_address = None

    if not other_address and not my_address:
        return

    msg_other += '\n'
    msg_other += '\n'
    msg_other += 'Ran by ' + username + '\n'
    msg_other += '\n'
    if critical_msgs:
        msg_other += 'Critical errors during the processing:\n'
        for m in critical_msgs:
            msg_other += '  ' + m + '\n'
        msg_other += '\n'

    if error_msgs:
        if critical_msgs:
            msg_other += 'Other e'
        else:
            msg_other += 'E'
        msg_other += 'rrors during the processing:\n'
        for m in error_msgs:
            msg_other += '  ' + m + '\n'

    msg_me = msg_other[:]
    if warning_msgs:
        msg_me += 'Warnings during the processing:\n'
        for m in warning_msgs:
            msg_me += '  ' + m + '\n'

    if not subj:
        subj = ''
        if project_name:
            subj += project_name
        else:
            subj += 'Reporting'
        if proc_name:
            subj += ' - ' + proc_name

    msg_other = MIMEText(msg_other)
    msg_me = MIMEText(msg_me)

    msg_other['Subject'] = msg_me['Subject'] = subj

    msg_other['From'] = msg_me['From'] = 'klpf990@rask.usbod.astrazeneca.com'
    msg_other['To'] = other_address
    msg_me['To'] = my_address

    def try_send(host, msg_):
        s = smtplib.SMTP(host)
        s.sendmail(msg_['From'], msg_['To'].split(','), msg_.as_string())
        s.quit()
        # info('Mail sent to ' + msg_['To'] + ' using ' + host)

    def print_msg():
        for line in msg_other.as_string().split('\n'):
            sys.stdout.write('   | ' + line + '\n')
        sys.stdout.write('\n')

    msgs = []
    if other_address:
        msgs.append(msg_other)
    if my_address:
        msgs.append(msg_me)
    for msg in msgs:
        try:
            try_send(smtp_host, msg)
        except smtplib.SMTPException:
            warn('Could not send email using the sever "' + smtp_host + '" with exception: ')
            warn('  ' + '; '.join(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])))
            if smtp_host != 'localhost':
                warn('Trying "localhost" as a server...')
                try:
                    try_send('localhost', msg)
                except smtplib.SMTPException:
                    warn('Could not send email using the sever "localhost" with exception: ')
                    warn('  ' + '; '.join(traceback.format_exception_only(sys.exc_info()[0], sys.exc_info()[1])))
                    print_msg()
            else:
                print_msg()


class CriticalError(Exception):
    pass


def _log(out, msg='', ending='\n', print_date=True, severity=None):
    msg_debug = str(msg)
    msg = str(msg)

    if print_date:
        msg_debug = timestamp() + '  ' + msg
        # if is_debug:
        #     msg_debug = severity + ' ' * (12 - len(severity)) + msg_debug

    if is_debug:
        out.write(msg_debug + ending)
    elif severity != 'debug':
        out.write(msg + ending)

    if severity == 'critical':
        critical_msgs.append(msg)
    if severity == 'error':
        error_msgs.append(msg)
    if severity == 'warning':
        warning_msgs.append(msg)

    sys.stdout.flush()
    sys.stderr.flush()
    if is_debug and is_local():
        sleep(0.01)

    if log_fpath:
        _write_to_file(msg_debug + ending)
    else:
        past_msgs.append(msg_debug + ending)


def _write_to_file(text):
    if log_fpath:
        try:
            open(log_fpath, 'a').write(text)
        except IOError:
            sys.stderr.write('Logging: cannot append to ' + log_fpath + '\n')
            try:
                open(log_fpath, 'w').write(text)
            except IOError:
                sys.stderr.write('Logging: cannot write to ' + log_fpath + '\n')


import os
from collections import OrderedDict
from genericpath import exists
from os.path import splitext, basename, join, isfile, isdir, abspath, realpath, islink
from random import random
import datetime

from ngs_utils import logger
from ngs_utils.bam_utils import verify_bam
from ngs_utils.file_utils import verify_file, adjust_path, splitext_plus, safe_mkdir, file_exists
from ngs_utils.logger import info, critical, err, debug


def read_samples(args):
    bam_by_sample = find_bams(args)
    if bam_by_sample:
        info('Found ' + str(len(bam_by_sample)) + ' BAM file' + ('s' if len(bam_by_sample) > 1 else ''))

    input_not_bam = [verify_file(fpath) for fpath in args if adjust_path(fpath) not in bam_by_sample]
    input_not_bam = [fpath for fpath in input_not_bam if fpath]
    fastqs_by_sample = dict()
    if not input_not_bam and not bam_by_sample:
        critical('No correct input files')
    if input_not_bam:
        info('Input ' + str(len(input_not_bam)) + ' correct input non-BAM files')
        fastqs_by_sample = find_fastq_pairs(input_not_bam)
        if fastqs_by_sample:
            info('Found ' + str(len(fastqs_by_sample)) + ' FastQ pairs')
        intersection = set(fastqs_by_sample.keys()) & set(bam_by_sample.keys())
        if intersection:
            critical('The following samples both had input BAMs and FastQ: ' + ', '.join(list(intersection)))

    return fastqs_by_sample, bam_by_sample


def find_bams(args):
    bam_by_sample = OrderedDict()
    bad_bam_fpaths = []

    good_args = []
    for arg in args:
        # /ngs/oncology/Analysis/bioscience/Bio_0038_KudosCellLinesExomes/Bio_0038_150521_D00443_0159_AHK2KTADXX/bcbio,Kudos159 /ngs/oncology/Analysis/bioscience/Bio_0038_KudosCellLinesExomes/Bio_0038_150521_D00443_0160_BHKWMNADXX/bcbio,Kudos160
        fpath = arg.split(',')[0]
        fname, ext = splitext(fpath)
        if ext == '.bam':
            bam_fpath = verify_bam(fpath)
            if not bam_fpath:
                bad_bam_fpaths.append(fpath)
            else:
                if len(arg.split(',')) > 1:
                    sname = arg.split(',')[1]
                else:
                    sname = basename(splitext(bam_fpath)[0])
                bam_by_sample[sname] = bam_fpath
                good_args.append(arg)
    if bad_bam_fpaths:
        critical('BAM files cannot be found, empty or not BAMs: ' + ', '.join(bad_bam_fpaths))
    for arg in good_args:
        args.remove(arg)

    return bam_by_sample


def find_fastq_pairs(fpaths):
    info('Finding FastQ pairs...')
    fastqs_by_sample_name = dict()
    for fpath in fpaths:
        fn, ext = splitext_plus(basename(fpath))
        if ext in ['.fq', '.fq.gz', '.fastq', '.fastq.gz']:
            sname, l_fpath, r_fpath = None, None, None
            if fn.endswith('_1'):
                sname = fn[:-2]
                l_fpath = fpath
            if fn.endswith('_R1'):
                sname = fn[:-3]
                l_fpath = fpath
            if fn.endswith('_2'):
                sname = fn[:-2]
                r_fpath = fpath
            if fn.endswith('_R2'):
                sname = fn[:-3]
                r_fpath = fpath
            if not sname:
                sname = fn
                info('Cannot detect file for ' + sname)

            l, r = fastqs_by_sample_name.get(sname, (None, None))
            if l and l_fpath:
                critical('Duplicated left FastQ files for ' + sname + ': ' + l + ' and ' + l_fpath)
            if r and r_fpath:
                critical('Duplicated right FastQ files for ' + sname + ': ' + r + ' and ' + r_fpath)
            fastqs_by_sample_name[sname] = l or l_fpath, r or r_fpath

    fixed_fastqs_by_sample_name = dict()
    for sname, (l, r) in fastqs_by_sample_name.items():
        if not l:
            err('ERROR: for sample ' + sname + ', left reads not found')
        if not r:
            err('ERROR: for sample ' + sname + ', right reads not found')
        if l and r:
            fixed_fastqs_by_sample_name[sname] = l, r

    return fixed_fastqs_by_sample_name


def set_up_dirs(proc_name, output_dir=None, work_dir=None, log_dir=None):
    """ Creates output_dir, work_dir, and sets up log
    """
    output_dir = safe_mkdir(adjust_path(output_dir or join(os.getcwd(), proc_name)), 'output_dir')
    debug('Saving results into ' + output_dir)

    work_dir = safe_mkdir(work_dir or join(output_dir, 'work'), 'working directory')
    info('Using work directory ' + work_dir)

    log_fpath = set_up_log(log_dir or work_dir, proc_name + '.log')

    return output_dir, work_dir, log_fpath


def set_up_log(log_dir, log_fname):
    log_fname = log_fname
    log_fpath = join(log_dir, log_fname)

    if file_exists(log_fpath):
        timestamp = datetime.datetime.fromtimestamp(os.stat(log_fpath).st_mtime)
        mv_log_fpath = log_fpath + '.' + timestamp.strftime('%Y-%m-%d_%H-%M-%S_' + str(random() * 1000))
        try:
            if isfile(mv_log_fpath):
                os.remove(mv_log_fpath)
            if not isfile(mv_log_fpath):
                os.rename(log_fpath, mv_log_fpath)
        except OSError:
            pass

    logger.set_log_path(log_fpath)
    debug('Logging to ' + log_fpath)
    debug()
    return log_fpath





#!/usr/bin/env python
'''
Merges tab-delimited files together, combining common columns
'''
import sys,os, gzip
from .support import filenames_to_uniq


class MergeException(Exception):
    pass
    
def merge_files(fnames,common_cols,uncommon_cols, keycols, noheader=False,collate=True,headercomment=False,keydesc=False,nomissing=False, given_names=None):
    if given_names:
        names = given_names
    else:
        names = filenames_to_uniq([os.path.basename(x) for x in fnames])
    files = []
    file_col_count = 0
    for fname in fnames:
        if fname[-3:] == '.gz':
            files.append(gzip.open(fname))
        else:
            files.append(open(fname))

    commented_lines = {}
    for name,f in zip(names,files):
        for line in f:
            if line[0] == '#':
                if not name in commented_lines:
                    commented_lines[name] = []
                    
                commented_lines[name].append(line[1:])
            else:
                break
        f.seek(0)
    
    if commented_lines:
        commentsout = False
        for name in commented_lines:
            for line in commented_lines[name][:-1]:
                commentsout = True
                sys.stdout.write('# %s %s' % (name,line))
            if not headercomment:
                commentsout = True
                sys.stdout.write('# %s %s' % (name,commented_lines[name][-1]))
                
        if commentsout:
            sys.stdout.write('\n')
    
    headers = True

    # header is the last commented line
    if headercomment:
        header_cols = []
        for name in commented_lines:
            cols = commented_lines[name][-1].rstrip().split('\t')
            for i in common_cols:
                header_cols.append(cols[i])
            if not collate:
                for j in uncommon_cols:
                    for name in names:
                        header_cols.append('%s %s' % (name,cols[j]))
            else:
                for name in names:
                    for j in uncommon_cols:
                        header_cols.append('%s %s' % (name,cols[j]))
            break
        sys.stdout.write('\t'.join(header_cols))
        sys.stdout.write('\n')
        headers = False
    
    #no header... just show column #
    
    elif noheader:
        header_cols = []
        for common in common_cols[1:]:
            header_cols.append('')
            sys.stdout.write('\t')
            
        if len(uncommon_cols) > 1:
            if not collate:
                for j in uncommon_cols:
                    for name in names:
                        header_cols.append('%s col %s' % (name,j))
            else:
                for name in names:
                    for j in uncommon_cols:
                        header_cols.append('%s col %s' % (name,j))
                
        else:
            for name in names:
                header_cols.append('%s' % (name))

        sys.stdout.write('\t'.join(header_cols))
        sys.stdout.write('\n')
        headers = False
        
    lines = []
    for f in files:
        lines.append(None)
        
    while True:
        missing_values = False
        
        for i,f in enumerate(files):
            try:
                while not lines[i]:
                    lines[i] = f.next()
                    
                    if lines[i][0] == '#' or lines[i].strip() == '':
                        lines[i] = None
                    
            except Exception as e:
                lines[i] = None
                pass
        
        good = False
        for line in lines:
            if line:
                good = True
        if not good:
            break
        
        outcols = []
        values = []
        common_keys = []
        num_of_columns = len(common_cols) + len(uncommon_cols)
        
        # look for missing values
        for i,line in enumerate(lines):
            if not line:
                missing_values = True
                values.append(['',] * num_of_columns)
                continue
                
            cols = line.rstrip().split('\t')
            if not cols:
                missing_values = True
                values.append(['',] * num_of_columns)
            
            while len(cols) < file_col_count:
                cols.append('')
                
            values.append(cols)
            
            keys = []
            for j,num in zip(keycols[0],keycols[1]):
                if num and not headers:
                    keys.append(float(cols[j]))
                else:
                    keys.append(cols[j])
            common_keys.append((keys,i))
            
        common_keys.sort()
        if keydesc:
            common_keys.reverse()
            
        for keys,i in common_keys:
            if keys == common_keys[0][0]:
                lines[i] = None
                if not outcols:
                    for j in common_cols:
                        outcols.append(values[i][j])
            else:
                missing_values = True
                values[i] = ['',] * num_of_columns

        if not values:
            continue

        if nomissing and missing_values:
            continue

        # first line is header
        if headers:
            headers = False
            if not file_col_count:
                file_col_count = len(values[0])
            if not collate:
                for j in uncommon_cols:
                    for name in names:
                        outcols.append('%s %s' % (name,values[0][j]))
            else:
                for name in names:
                    for j in uncommon_cols:
                        outcols.append('%s %s' % (name,values[0][j]))
        else:
            if not collate:
                for j in uncommon_cols:
                    for cols in values:
                        if len(cols)>j:
                            outcols.append(cols[j])
                        else:
                            outcols.append('')

            else:
                for cols in values:
                    for j in uncommon_cols:
                        if len(cols)>j:
                            outcols.append(cols[j])
                        else:
                            outcols.append('')
        if outcols:
            sys.stdout.write('%s\n' % '\t'.join(outcols))

    
    for f in files:
        f.close()
        
def usage():
    print(__doc__)
    print("""\
Usage: %s {opts} common_cols merge_cols files

common_cols and merge_cols should be a comma-separated list of column numbers.

Files must be in the same sort order (given as  if not ascending text).
If the case of extra rows in one of the files, blank values will be 
substituted. New column names will be guessed based upon the filenames. 
Commented lines and blank lines are ignored, except for any commented lines 
that are at the begining of the file - these are kept.  This assumes that the 
first non-comment, non-blank row is the header.

Options:
    -headercomment     the header is last commented line ('#')
    -noheader          the files have no header row
    -nomissing         discard rows with missing values
    -collate           order uncommon values by file first, not column
    
    -keycols col,col   if there are missing values, use these columns to 
                       determine which file has missing data.  If the col ends 
                       in 'n', this is taken to be a number.
                       (defaults to common cols, in order)
                       
    -keydesc           keys are sorted descending order
                       (defaults to ascending)

Valid column definitions:
    1,2,3,4
    1-4
    1-4,5

""" % os.path.basename(sys.argv[0]))

def _split_cols(arg):
    ret=[]
    nums=[]
    for x in arg.split(','):
        num = False
        if x[-1] == 'n':
            num = True
            x = x[:-1]
        
        if '-' in x:
            s,e = x.split('-')
            vals = list(range(int(s)-1,int(e)))
            ret.extend(vals)
            for v in vals:
                nums.append(num)
        else:
            ret.append(int(x)-1)
            nums.append(num)
    return ret,nums

def main(argv):
    noheader=False
    collate = False
    common = None
    uncommon = None
    keycols = None
    keydesc = False
    headercomment = False
    nomissing = False
    files=[]
    given_names=[]
    
    last = None
    
    for arg in argv:
        if arg == '-h':
            usage()
            sys.exit(1)
        if last == '-keycols':
            keycols = _split_cols(arg)
            last = None
        elif last == '-names':
            given_names = [x.strip() for x in arg.split(',')]
            last = None
        elif arg in ['-keycols', '-names']:
            last = arg
        elif arg == '-keydesc':
            keydesc = True
        elif arg == '-nomissing':
            nomissing = True
        elif arg == '-headercomment':
            headercomment = True
        elif arg == '-noheader':
            noheader = True
        elif arg == '-collate':
            collate = True
        elif not common:
            common = _split_cols(arg)
        elif not uncommon:
            uncommon = _split_cols(arg)
        elif os.path.exists(arg):
            files.append(arg)
        else:
            usage()
            sys.exit(1)
    if not files or not common or not uncommon:
        usage()
        sys.exit(1)
    
    if not keycols:
        keycols = common
    
    merge_files(files,common[0],uncommon[0],keycols,noheader,collate,headercomment,keydesc,nomissing, given_names)

if __name__ == '__main__':
    main(sys.argv[1:])



from contextlib import contextmanager
from traceback import format_exc

from ngs_utils.logger import info, err, critical, debug, warn
from ngs_utils.file_utils import verify_file, verify_module, adjust_path

from yaml import load as load_yaml


def load_yaml_config(fpath):
    verify_file(fpath, is_critical=True)
    try:
        dic = load_yaml(open(fpath))
    except Exception:
        err(format_exc())
        critical('Could not parse bcbio YAML ' + fpath)
    else:
        return dic


def fill_dict_from_defaults(cur_cnf, defaults_dict):
    for key in defaults_dict:
        if key in cur_cnf:
            if isinstance(cur_cnf[key], dict) and isinstance(defaults_dict[key], dict):
                fill_dict_from_defaults(cur_cnf[key], defaults_dict[key])
        else:
            cur_cnf[key] = defaults_dict[key]
    return cur_cnf


def _join_parent_conf(child_conf, parent_conf):
    bc = parent_conf.copy()
    bc.update(child_conf)
    child_conf.update(bc)
    return child_conf


@contextmanager
def with_cnf(cnf, **kwargs):
    prev_opts = {k: cnf[k] for k in kwargs.keys()}
    try:
        for k, v in kwargs.items():
            cnf[k] = v
        yield cnf
    finally:
        for k, v in prev_opts.items():
            cnf[k] = v



import subprocess
import unittest
import os
from os.path import dirname, join, exists, isfile, splitext, basename, isdir, relpath, getctime, getsize, abspath, expanduser, \
    realpath
import shutil
import sys
from datetime import datetime

from ngs_utils.file_utils import verify_dir, verify_file


def info(msg=''):
    sys.stdout.write(msg + '\n')

def call(cmdl):
    info(cmdl if isinstance(cmdl, str) else subprocess.list2cmdline(cmdl))
    if isinstance(cmdl, str):
        return subprocess.call(cmdl, shell=True, executable='/bin/bash')
    else:
        return subprocess.call(cmdl)

def check_call(cmdl):
    info(cmdl if isinstance(cmdl, str) else subprocess.list2cmdline(cmdl))
    if isinstance(cmdl, str):
        subprocess.check_call(cmdl, shell=True, executable='/bin/bash')
    else:
        subprocess.check_call(cmdl)

def check_output(cmdl):
    info(cmdl if isinstance(cmdl, str) else subprocess.list2cmdline(cmdl))
    if isinstance(cmdl, str):
        return subprocess.check_output(cmdl, shell=True, executable='/bin/bash', stderr=subprocess.STDOUT)
    else:
        return subprocess.check_output(cmdl, stderr=subprocess.STDOUT)

def swap_output(output_path):
    if not exists(output_path):
        return

    last_changed = datetime.fromtimestamp(getctime(output_path))
    prev_output_path = output_path + '_' + last_changed.strftime('%Y_%m_%d__%H_%M_%S')
    os.rename(output_path, prev_output_path)
    return prev_output_path

def swap_prev_symlink(output_path, prev_output_path):
    prev_output_link = get_prev(output_path)
    if exists(prev_output_link):
        os.remove(prev_output_link)
    if prev_output_path:
        os.symlink(prev_output_path, prev_output_link)

def get_prev(fpath):
    return fpath + '_prev'


class BaseTestCase(unittest.TestCase):
    script = None

    data_dir = 'data'
    results_dir = 'results'
    gold_standard_dir = 'gold_standard'

    remove_work_dir_on_success = False

    def setUp(self):
        if not isdir(self.data_dir):
            os.makedirs(self.data_dir)
        if not exists(self.results_dir):
            os.makedirs(self.results_dir)

    def _check_file_throws(self, fpath, ignore_matching_lines=None, wrapper=None, cmp_line_number_only=True, check_diff=True):
        assert isfile(fpath), 'file does not exist: ' + fpath
        assert getsize(fpath) > 0, 'file is empty: ' + fpath

        if check_diff:
            cmp_fpath = None
            if isdir(self.gold_standard_dir):
                cmp_fpath = join(self.gold_standard_dir, relpath(fpath, self.results_dir))
            elif isfile(get_prev(fpath)):
                cmp_fpath = get_prev(fpath)

            if cmp_fpath and isfile(cmp_fpath):
                cmdl = 'diff'
                if ignore_matching_lines:
                    if isinstance(ignore_matching_lines, str):
                        ignore_matching_lines = [ignore_matching_lines]
                    for r in ignore_matching_lines:
                        cmdl += ' -I ' + subprocess.list2cmdline([r])
                if wrapper:
                    if isinstance(ignore_matching_lines, str):
                        wrapper.split()
                    wrapper = subprocess.list2cmdline(wrapper)
                    if not fpath.endswith('.gz'):
                        fpath = '<(' + wrapper + ' ' + fpath + ')'
                        cmp_fpath = '<(' + wrapper + ' ' + cmp_fpath + ')'
                    else:
                        fpath = '<(gunzip -c ' + fpath + ' | ' + wrapper + ')'
                        cmp_fpath = '<(gunzip -c ' + cmp_fpath + ' | ' + wrapper + ')'
                elif fpath.endswith('.gz'):
                    fpath = '<(gunzip -c ' + fpath + ')'
                    cmp_fpath = '<(gunzip -c ' + cmp_fpath + ')'
                cmdl += ' ' + fpath + ' ' + cmp_fpath
                ret_code = call(cmdl)
                assert ret_code == 0, 'diff returned non-zero: ' + fpath

    @staticmethod
    def _check_dir_not_empty(dirpath, description=None):
        assert verify_dir(dirpath, description=description), dirpath
        contents = [join(dirpath, fname) for fname in os.listdir(dirpath)
                    if not fname.startswith('.')]
        assert len(contents) >= 1, dirpath + ': ' + str(contents)
        assert all(verify_file(realpath(fpath), is_critical=True)
                   for fpath in contents
                   if isfile(realpath(fpath))), dirpath + ': ' + str(contents)



#!/usr/bin/env python
import sys
from collections import defaultdict

from Utils.tools import read_approved_genes, get_approved_gene_symbol


def main():
    if len(sys.argv) < 2:
        log('Usage: ' + __file__ + ' genes.txt [regions.bed] [HGNC_gene_synonyms.txt] [--not-check] > regions_for_genes.txt')
        log('    the script filters regions to have gene names in genes.txt')
        sys.exit(1)

    genes_list_fpath = sys.argv[1]
    exons_fpath = '/ngs/reference_data/genomes/Hsapiens/hg19/bed/Exons/RefSeq/RefSeq_CDS.hg19.bed'
    synonyms_fpath = '/ngs/reference_data/genomes/Hsapiens/common/HGNC_gene_synonyms.txt'
    if len(sys.argv) > 2: exons_fpath = sys.argv[2]
    if len(sys.argv) > 3: synonyms_fpath = sys.argv[3]
    not_check = len(sys.argv) > 4

    approved_gene_by_name, approved_gnames_by_prev_gname, approved_gnames_by_synonym = \
        read_approved_genes(synonyms_fpath)

    gnames = set()
    duplicated_genes = list()
    with open(genes_list_fpath) as f:
        for i, l in enumerate(f):
            gname = l.strip()
            sys.stderr.write(str(i + 1) + ' ' + gname + ' ')

            if gname in gnames:
                duplicated_genes.append(gname)
                sys.stderr.write(' duplicated gene\n')
                continue

            gnames.add(gname)
            sys.stderr.write('\n')

    log('Total unique gene names: ' + str(len(gnames)))
    if duplicated_genes:
        log(str(len(duplicated_genes)) + ' duplicated: ' + ', '.join(duplicated_genes))
    log()

    genes_found_in_ref, lines = exons_for_gene_list(exons_fpath, gnames)
    # log(str(len(doublicated_after_approving)) + ' duplicated after approval: ' + ', '.join(doublicated_after_approving))
    # log('Not found in HGNC ' + str(len(not_found_in_HGNC_genes)) + ' genes' +
    #     (': ' + ', '.join(not_found_in_HGNC_genes) if not_found_in_HGNC_genes else ''))
    log('Found in Exons DB ' + str(len(genes_found_in_ref)) + ' genes out of ' + str(len(gnames)))
    not_found_in_reference = gnames - genes_found_in_ref
    log('Not found in Exons DB ' + str(len(not_found_in_reference)) + ': ' + ', '.join(not_found_in_reference))

    doublicated_after_approving_gnames = list()
    doublicated_after_approving_approved = list()
    original_by_approved = defaultdict(list)
    corrected_genes = set()
    corrected_not_found_in_HGNC_genes = set()
    not_corrected = set()
    if not_found_in_reference:
        log()
        log('Correcting genes not found in Exons DB')
        for gname in not_found_in_reference:
            log(gname)

            approved_gname, status = get_approved_gene_symbol(
                approved_gene_by_name, approved_gnames_by_prev_gname, approved_gnames_by_synonym, gname,
                indent='    ')

            if not approved_gname:
                if '.' in gname:
                    gname2 = gname.split('.')[0]
                    log('    Not found with dot, trying without dot, as ' + gname2 + '...')
                    approved_gname, status2 = get_approved_gene_symbol(
                        approved_gene_by_name, approved_gnames_by_prev_gname, approved_gnames_by_synonym, gname2,
                        indent='    ')

            if approved_gname:
                if gname == approved_gname:
                    log('    Gene name ' + gname + ' is already approved. Skipping.')
                    not_corrected.add(gname)
                else:
                    log('    Found as ' + approved_gname)

                    if approved_gname in gnames:
                        log('    Approved version for ' + gname + ' is already met as ' + approved_gname +
                                         ' in gnames')
                        original_by_approved[approved_gname].append(gname)
                        doublicated_after_approving_gnames.append(gname + '->' + approved_gname)
                        continue

                    if approved_gname in original_by_approved:
                        log('    Approved version for ' + gname + ' is already met as ' + approved_gname +
                                         ' for ' + ', '.join(original_by_approved[approved_gname]))
                        original_by_approved[approved_gname].append(gname)
                        doublicated_after_approving_approved.append(gname + '->' + approved_gname)
                        continue

                    log('    Found unique corrected approved gene name ' + approved_gname)
                    original_by_approved[approved_gname].append(gname)
                    corrected_genes.add(approved_gname)

            else:
                log('    ' + gname + ' not found in HGNC.')
                corrected_not_found_in_HGNC_genes.add(gname)
            log()

    log('-' * 70)
    log('Total unique gene names: ' + str(len(gnames)))
    if duplicated_genes:
        log('    ' + str(len(duplicated_genes)) + ' duplicated: ' + ', '.join(duplicated_genes))

    log()
    log('Searching exons for the gene list: first iteration')
    log('    Found ' + str(len(lines)) + ' exons for ' + str(len(genes_found_in_ref)) + ' genes (out of ' + str(len(gnames)) + ')')
    not_found_in_reference = gnames - genes_found_in_ref
    log('    Not found in Exons DB ' + str(len(not_found_in_reference)) + ': ' + ', '.join(not_found_in_reference))

    corrected_genes_found_in_ref = set()
    corrected_not_found_in_reference = set()
    if not_found_in_reference:
        log()
        log('Checking failed gene names with HGNC')
        log('    ' + str(len(doublicated_after_approving_gnames)) + ' approved names are already accounted for in the first iteration: ' + ', '.join(doublicated_after_approving_gnames))
        log('    ' + str(len(doublicated_after_approving_approved)) + ' duplicated approved names' + ((': ' + ', '.join(doublicated_after_approving_approved)) if doublicated_after_approving_approved else ''))
        log('    ' + str(len(corrected_not_found_in_HGNC_genes)) + ' not found in HGNC' + (': ' + ', '.join(corrected_not_found_in_HGNC_genes) if corrected_not_found_in_HGNC_genes else ''))
        log('    ' + str(sum([len(original_by_approved[c]) for c in corrected_genes])) + ' gene names corrected by HGNC into new names: ' + ', '.join([':'.join(original_by_approved[c]) + '->' + c for c in corrected_genes]))
        log('    ' + str(len(not_corrected)) + ' not corrected gene names.')

        log('Writing exons for the approved version of gene names originally not found in the Exons DB')
        corrected_genes_found_in_ref, lines_2 = exons_for_gene_list(exons_fpath, corrected_genes)
        corrected_not_found_in_reference = corrected_genes - corrected_genes_found_in_ref
        log('    Found ' + str(len(lines_2)) + ' new exons for ' + str(len(corrected_genes_found_in_ref)) + ' genes (out of ' + str(len(corrected_genes)) + ')')
        log('    Out of corrected, not found in Exons DB ' + str(len(corrected_not_found_in_reference)) + ((': ' + ', '.join(corrected_not_found_in_reference)) if corrected_not_found_in_reference else ''))

    log('')
    total_uniq_genes_found = len(gnames) - len(corrected_not_found_in_HGNC_genes) - len(corrected_not_found_in_reference)
    log('Finally writing exons for ' + str(total_uniq_genes_found) + ' unique gene name records, ' +
        str(len(genes_found_in_ref | corrected_genes_found_in_ref)) + ' unique genes after correcting with HGNC')
    genes_found_in_ref, lines = exons_for_gene_list(exons_fpath, genes_found_in_ref | corrected_genes_found_in_ref)
    log('    Written ' + str(len(lines)) + ' lines for ' + str(len(genes_found_in_ref)) + ' genes')
    missed_genes = not_corrected | corrected_not_found_in_HGNC_genes
    log('    Missed ' + str(len(missed_genes)) + ' genes: ' + ', '.join(missed_genes))

    for l in lines:
        print(l)


def exons_for_gene_list(exons_fpath, gnames):
    genes_found_in_ref = set()
    lines = []

    with open(exons_fpath) as f:
        for l in f:
            if not l.startswith('#'):
                fs = l.strip().split('\t')
                if len(fs) < 7:
                    pass
                else:
                    feature = fs[6]
                    # if feature in ['Gene', 'Multi_Gene']:
                    #     pass
                    # else:
                    gname = fs[3]
                    if gname not in gnames:
                        pass
                        # genes_not_found_in_ref.add(gname)
                    else:
                        genes_found_in_ref.add(gname)
                        lines.append('\t'.join(fs))

    return genes_found_in_ref, lines


def log(msg=''):
    sys.stderr.write(msg + '\n')


if __name__ == '__main__':
    main()



import os
import sys
import math
from collections import OrderedDict

import pybedtools
from ngs_utils.call_process import run
from os.path import isfile, join, abspath, basename, dirname, getctime, getmtime, splitext, realpath
from subprocess import check_output
from pybedtools import BedTool

from ngs_utils import call_process
from ngs_utils.file_utils import intermediate_fname, iterate_file, splitext_plus, verify_file, adjust_path, add_suffix, \
    safe_mkdir, file_transaction, which, file_exists, open_gzipsafe, can_reuse
from ngs_utils.logger import info, critical, warn, err, debug
from ngs_utils import reference_data as ref
from ngs_utils.utils import md5


def get_chrom_order(genome=None, fai_fpath=None):
    chr_lengths = ref.get_chrom_lengths(genome, fai_fpath)
    chr_order = {c: i for i, (c, l) in enumerate(chr_lengths)}
    return chr_order


class SortableByChrom:
    def __init__(self, chrom, chrom_ref_order):
        self.chrom = chrom
        if isinstance(chrom_ref_order, dict):
            self.chrom_ref_order = chrom_ref_order.get(chrom, -1)
        else:
            self.chrom_ref_order = chrom_ref_order

    def get_key(self):
        return self.chrom_ref_order


class Region(SortableByChrom):
    def __init__(self, chrom, start, end, other_fields, chrom_ref_order):
        SortableByChrom.__init__(self, chrom, chrom_ref_order)
        self.start = int(start)
        self.end = int(end) if end is not None else None
        self.other_fields = list(other_fields)

    def get_key(self):
        return self.chrom_ref_order, self.start, self.end, self.other_fields

    def __str__(self):
        ts = [self.chrom, self.start, self.end] + self.other_fields
        return '"' + '\t'.join(map(str, ts)) + '"'

    def __repr__(self):
        return self.__str__()


def bam_to_bed(bam_fpath, to_gzip=True):
    debug('Converting the BAM to BED to save some memory.')  # from here: http://davetang.org/muse/2015/08/05/creating-a-coverage-plot-using-bedtools-and-r/
    bam_bed_fpath = splitext_plus(bam_fpath)[0] + ('.bed.gz' if to_gzip else '.bed')
    if can_reuse(bam_bed_fpath, bam_fpath):
        return bam_bed_fpath
    bedtools = which('bedtools')
    gzip = which('gzip')
    cmdline = '{bedtools} bamtobed -i {bam_fpath}'.format(**locals())
    cmdline += ' | {gzip}'.format(**locals()) if to_gzip else ''
    call_process.run(cmdline, output_fpath=bam_bed_fpath)
    return bam_bed_fpath


def bam_to_bed_nocnf(bam_fpath, bedtools='bedtools', gzip='gzip'):
    info('Converting the BAM to BED to save some memory.')  # from here: http://davetang.org/muse/2015/08/05/creating-a-coverage-plot-using-bedtools-and-r/
    bam_bed_fpath = splitext_plus(bam_fpath)[0] + '.bed.gz'
    cmdline = '{bedtools} bamtobed -i {bam_fpath} | {gzip} > {bam_bed_fpath}'.format(**locals())
    info(cmdline)
    os.system(cmdline)
    bam_bed_fpath = verify_file(bam_bed_fpath)
    if bam_bed_fpath:
        info('Done, saved to ' + bam_bed_fpath)
    else:
        err('Error, result is non-existent or empty')
    return bam_bed_fpath


def count_bed_cols(bed_fpath):
    with open(bed_fpath) as f:
        for l in f:
            if l and l.strip() and not l.startswith('#'):
                return len(l.split('\t'))
    # return len(next(dropwhile(lambda x: x.strip().startswith('#'), open(bed_fpath))).split('\t'))
    err('Empty bed file: ' + bed_fpath)
    return None


def merge_overlaps(work_dir, bed_fpath, distance=None):
    """Merge bed file intervals to avoid overlapping regions.
    Overlapping regions (1:1-100, 1:90-100) cause issues with callers like FreeBayes
    that don't collapse BEDs prior to using them.
    """
    output_fpath = intermediate_fname(work_dir, bed_fpath, 'merged')
    if isfile(output_fpath) and verify_file(output_fpath, cmp_f=bed_fpath):
        return output_fpath

    with file_transaction(work_dir, output_fpath) as tx:
        kwargs = dict(d=distance) if distance else dict()
        BedTool(bed_fpath).merge(**kwargs).saveas(tx)
    return output_fpath


def calc_region_number(bed_fpath):
    with open(bed_fpath) as f:
        return sum(1 for l in f if l.strip() and not l.strip().startswith('#'))


def get_genes_from_bed(bed_fpath, chrom_index=0, gene_index=3):
    gene_keys_set = set()
    gene_keys_list = list()
    with open(bed_fpath) as f:
        for line in f:
            if not line or not line.strip() or line.startswith('#'):
                continue

            tokens = line.replace('\n', '').split('\t')
            if len(tokens) <= gene_index or len(tokens) <= chrom_index:
                continue

            chrom = tokens[chrom_index]
            for gn in tokens[gene_index].split(','):
                if (gn, chrom) not in gene_keys_set:
                    gene_keys_set.add((gn, chrom))
                    gene_keys_list.append((gn, chrom))

    return gene_keys_set, gene_keys_list


def cut(fpath, col_num, output_fpath=None):
    output_fpath = output_fpath or add_suffix(fpath, 'cut')
    if can_reuse(output_fpath, fpath):
        return output_fpath
    cmdline = 'cut -f' + ','.join(map(str, range(1, col_num + 1))) + ' ' + fpath + ' > ' + output_fpath
    call_process.run(cmdline, output_fpath=output_fpath)
    return output_fpath


def bgzip_and_tabix(fpath, reuse=False, tabix_parameters='', **kwargs):
    gzipped_fpath = join(fpath + '.gz')
    tbi_fpath = gzipped_fpath + '.tbi'

    if reuse and \
           file_exists(gzipped_fpath) and (getctime(gzipped_fpath) >= getctime(fpath) if file_exists(fpath) else True) and \
           file_exists(tbi_fpath) and getctime(tbi_fpath) >= getctime(gzipped_fpath):
        info('Actual compressed file and index exist, reusing')
        return gzipped_fpath

    info('Compressing and tabixing file, writing ' + gzipped_fpath + '(.tbi)')
    bgzip = which('bgzip')
    tabix = which('tabix')
    if not bgzip:
        err('Cannot index file because bgzip is not found')
    if not tabix:
        err('Cannot index file because tabix is not found')
    if not bgzip and not tabix:
        return fpath

    if isfile(gzipped_fpath):
        os.remove(gzipped_fpath)
    if isfile(tbi_fpath):
        os.remove(tbi_fpath)

    info('BGzipping ' + fpath)
    cmdline = '{bgzip} {fpath}'.format(**locals())
    call_process.run(cmdline)

    info('Tabixing ' + gzipped_fpath)
    cmdline = '{tabix} {tabix_parameters} {gzipped_fpath}'.format(**locals())
    call_process.run(cmdline)

    return gzipped_fpath


def filter_bed_with_gene_set(bed_fpath, gene_keys_set, output_fpath):
    with file_transaction(None, output_fpath) as tx:
        with open(bed_fpath) as inp, open(tx, 'w') as out:
            for l in inp:
                if l.strip('\n'):
                    chrom, start, end, gene = l.strip('\n').split('\t')
                    if (gene, chrom) in gene_keys_set:
                        out.write(l)


def sort_bed(input_bed_fpath, output_bed_fpath=None, work_dir=None, fai_fpath=None, chr_order=None, genome=None):
    input_bed_fpath = verify_bed(input_bed_fpath, is_critical=True)
    output_bed_fpath = adjust_path(output_bed_fpath) if output_bed_fpath \
        else intermediate_fname(work_dir, input_bed_fpath, 'sorted')

    debug('Sorting regions in ' + str(input_bed_fpath))
    if can_reuse(output_bed_fpath, input_bed_fpath):
        debug(output_bed_fpath + ' exists, reusing')
        return output_bed_fpath

    regions = []

    if not chr_order:
        if fai_fpath:
            fai_fpath = verify_file(fai_fpath)
        elif genome:
            fai_fpath = verify_file(ref.get_fai(genome))
        else:
            critical('Either of chr_order, fai_fpath, or genome build name must be specified')
        chr_order = get_chrom_order(fai_fpath=fai_fpath)

    with open(input_bed_fpath) as f:
        with file_transaction(work_dir, output_bed_fpath) as tx:
            with open(tx, 'w') as out:
                for l in f:
                    if not l.strip():
                        continue
                    if l.strip().startswith('#'):
                        out.write(l)
                        continue

                    fs = l.strip().split('\t')
                    chrom = fs[0]
                    start = int(fs[1])
                    end = int(fs[2])
                    other_fields = fs[3:]
                    order = chr_order.get(chrom, -1)
                    regions.append(Region(chrom, start, end, other_fields, order))

                for region in sorted(regions, key=lambda r: r.get_key()):
                    fs = [region.chrom, str(region.start), str(region.end)]
                    fs.extend(region.other_fields)
                    out.write('\t'.join(fs) + '\n')

    debug('Sorted ' + str(len(regions)) + ' regions, saved to ' + output_bed_fpath)
    return output_bed_fpath


def annotate_target(work_dir, target_bed, genome_build):
    output_fpath = intermediate_fname(work_dir, target_bed, 'ann')
    if can_reuse(output_fpath, target_bed):
        info(output_fpath + ' exists, reusing')
        return output_fpath

    annotate_bed_py = which('annotate_bed.py')
    if not annotate_bed_py:
        critical('Error: annotate_bed.py not found in PATH, please install TargQC.')

    cmdline = '{annotate_bed_py} {target_bed} -g {genome_build} -o {output_fpath}'.format(**locals())
    run(cmdline, output_fpath, stdout_to_outputfile=False)
    output_fpath = clean_bed(output_fpath, work_dir)
    return output_fpath


def calc_sum_of_regions(bed_fpath):
    total_bed_size = 0

    with open(bed_fpath) as f:
        for l in f:
            l = l.strip()
            if not l.startswith('#'):
                start, end = [int(f) for f in l.split('\t')[1:3]]
                total_bed_size += end - start

    return total_bed_size


def get_total_bed_size(bed_fpath, work_dir=None):
    if work_dir:
        pybedtools.set_tempdir(safe_mkdir(join(work_dir, 'pybedtools_tmp')))
    return sum(len(x) for x in BedTool(bed_fpath).merge())


def bedtools_version(bedtools):
    v = check_output([bedtools, '--version'])  # bedtools v2.24.0
    try:
        v = int(v.split(' ')[1].split('.')[1])
    except Exception:
        return None
    else:
        return v


this_py_fpath = splitext(__file__)[0] + '.py'
this_py_real_fpath = realpath(this_py_fpath)
project_dir = dirname(this_py_real_fpath)
code_base_path = abspath(join(abspath(project_dir)))


def get_padded_bed_file(work_dir, bed, padding, fai_fpath):
    genome_fpath = fai_fpath
    info('Making bed file for padded regions...')
    bedtools = which('bedtools')
    cmdline = '{bedtools} slop -i {bed} -g {genome_fpath} -b {padding}'.format(**locals())
    output_fpath = intermediate_fname(work_dir, bed, 'padded')
    call_process.run(cmdline, output_fpath=output_fpath)
    return output_fpath


def intersect_bed(work_dir, bed1, bed2):
    bed1_fname, _ = splitext_plus(basename(bed1))
    bed2_fname, _ = splitext_plus(basename(bed2))
    output_fpath = join(work_dir, bed1_fname + '__' + bed2_fname + '.bed')
    if can_reuse(output_fpath, [bed1, bed2]):
        return output_fpath
    bedtools = which('bedtools')
    cmdline = '{bedtools} intersect -u -a {bed1} -b {bed2}'.format(**locals())
    call_process.run(cmdline, output_fpath=output_fpath, checks=[call_process.file_exists_check])
    return output_fpath


def verify_bed(bed, description='', is_critical=False, silent=False):
    if isinstance(bed, BedTool):
        return bed

    fpath = adjust_path(bed)
    if not verify_file(fpath, description, is_critical=is_critical, silent=silent):
        return None

    error = BedFile(fpath).checkformat()
    if error:
        fn = critical if is_critical else err
        fn('Error: incorrect bed file format (' + fpath + '): ' + str(error) + '\n')
        return None

    return fpath


def clean_bed(bed_fpath, work_dir):
    clean_fpath = intermediate_fname(work_dir, bed_fpath, 'clean')

    if not can_reuse(clean_fpath, bed_fpath):
        pybedtools.set_tempdir(safe_mkdir(join(work_dir, 'pybedtools_tmp')))
        bed = BedTool(bed_fpath)
        bed = bed.filter(lambda x: x.chrom and
                         not any(x.chrom.startswith(e) for e in ['#', ' ', 'track', 'browser']))
        bed = bed.remove_invalid()
        with file_transaction(work_dir, clean_fpath) as tx_out_file:
            bed.saveas(tx_out_file)
        verify_bed(clean_fpath, is_critical=True)
        debug('Saved clean BED file into ' + clean_fpath)
    return clean_fpath


def check_md5(work_dir, fpath, file_type, silent=False):
    md5_fpath = join(work_dir, file_type + '_md5.txt')
    new_md5 = md5(fpath)
    info('md5 of ' + fpath + ' is ' + str(new_md5))
    prev_md5 = None
    if isfile(md5_fpath):
        with open(md5_fpath) as f:
            prev_md5 = f.read()
    else:
        info('Previous md5 file ' + md5_fpath + ' does not exist')
    info('Checking previous md5 from ' + md5_fpath + ': ' + str(prev_md5))

    if prev_md5 == new_md5:
        if not silent:
            debug('Reusing previous ' + file_type.upper() + ' files.')
        return True
    else:
        if not silent:
            info('Pre-processing input ' + file_type.upper() + ' file')
        if prev_md5:
            if not silent:
                info('Prev ' + file_type.upper() + ' md5: ' + str(prev_md5))
                info('New ' + file_type.upper() + ' md5: ' + str(new_md5))

        with open(md5_fpath, 'w') as f:
            f.write(str(new_md5))
        return False


class BedFile:
    def __init__(self, _filename):
        self.filename = _filename
        self.chrs = None
        self.nregions = None

    def checkformat(self):
        """************************************************************************************************************************************************************
        Task: checks the format of the bed file. The only requirements checked are that each line presents at least 3 tab separated columns, the
            two on the right must present integer values indicating the start/end position respectively. Right value must be greater than the
            left value.
        Outputs:
            err: string containing the detected error. Empty string in case of a correct format.
        ************************************************************************************************************************************************************"""

        fd = open_gzipsafe(self.filename)

        line = fd.readline()
        while line.startswith('#'):
            line = fd.readline()
        
        fields = line.split('\t')
        lc = 1
        error = ''

        # Checks that the two columns on the right contain integer values
        try:
            # Parses each line and checks that there are at least 3 fields, the two on the right containing integer values and being the right one
            # greater than the left one
            while line != '' and len(fields) > 2 and int(fields[1]) <= int(fields[2]):
                lc += 1
                line = fd.readline()
                fields = line.split('\t')
        except ValueError:
            error += 'Incorrect start/end values at line ' + str(lc) + '\n'
            error += 'Start/End coordinates must be indicated with integer values. The right value must be greater than the left value.\n'
            error += 'Line found: ' + line
            fd.close()

            return error

        # If it get to this point means that either the file ended or there is a line with less than 3 fields
        if line != '':
            error += 'Incorrect line format at line ' + str(lc) + '\n'
            error += 'At least three columns are expected in each line\n'
            error += 'The right value must be greater than the left value.\n'
            error += 'Line found: ' + line
            fd.close()

        return error

    def count_lines(self, filename=None):
        if filename is None:
            filename = self.filename
        return len(open(filename).readlines())


def calc_rate_within_normal(bases_by_depth, avg_depth, total_size):
    bases_within_normal = sum(
        bases
        for depth, bases
        in bases_by_depth.items()
        if math.fabs(avg_depth - depth) < 0.2 * avg_depth)

    return 1.0 * bases_within_normal / total_size if total_size else None


def calc_bases_within_threshs(bases_by_depth, total_size, depth_thresholds):
    bases_within_threshs = OrderedDict((depth, 0) for depth in depth_thresholds)
    rates_within_threshs = OrderedDict((depth, None) for depth in depth_thresholds)

    for depth, bases in bases_by_depth.items():
        for t in depth_thresholds:
            if depth >= t:
                bases_within_threshs[t] += bases
    for t in depth_thresholds:
        bs = bases_within_threshs[t]
        if total_size > 0:
            rate = 1.0 * bases_within_threshs[t] / total_size
            if rate > 1:
                critical('Error: rate is > 1: rate = ' + str(rate) + ', bases = ' + str(bs) + ', size = ' + str(total_size))
            rates_within_threshs[t] = rate

    return bases_within_threshs, rates_within_threshs



#!/usr/bin/env python
import os
import sys
from os.path import join, isfile, abspath, dirname, relpath, isdir
import shutil


def err(msg=''):
    sys.stderr.write(msg + '\n')
    

def init(name, package_name, setup_py_fpath, kwargs=None):
    if abspath(dirname(setup_py_fpath)) != abspath(os.getcwd()):
        sys.exit('Please, change to ' + dirname(setup_py_fpath) + ' before running setup.py\n')

    if len(sys.argv) == 2:
        cmd = sys.argv[1]
        if cmd == 'tag':
            version = write_version_py(package_name, kwargs=kwargs)
            run_cmdl('git tag -a %s -m "Version %s"' % (version, version))
            run_cmdl('git push --tags')
            sys.exit()
    
        if cmd == 'publish':
            run_cmdl('python setup.py sdist upload')
            sys.exit()
    
        if cmd == 'up':
            run_cmdl('git pull --recurse-submodules --rebase')
            # if first time: $ git submodule update --init --recursive
            run_cmdl('git submodule foreach "(git checkout master; git pull --rebase)"')
            sys.exit()
    
        if is_cleaning():
            clean_package(package_name, '.')
            sys.exit()

    # print('Upgrading pip and setuptools...')
    # try:
    #     os.subprocess.call('pip install --upgrade pip', shell=True)
    #     os.subprocess.call('pip install --upgrade --ignore-installed setuptools', shell=True)
    # except Exception:
    #     err('Cannot update pip and setuptools, that might cause errors during the following installation')

    version = write_version_py(package_name, kwargs=kwargs)
    print('Installing ' + name + ((' version ' + str(version)) if version else ''))
    print('')
    return version


def clean_package(package_name, dirpath='.'):
    print('Cleaning up binary, build and dist for ' + package_name + ' in ' + dirpath + '...')
    if isdir(join(dirpath, 'build')):
        shutil.rmtree(join(dirpath, 'build'))
    if isdir(join(dirpath, 'dist')):
        shutil.rmtree(join(dirpath, 'dist'))
    if isdir(join(dirpath, package_name + '.egg-info')):
        shutil.rmtree(join(dirpath, package_name + '.egg-info'))
    print('Done.')


def get_reqs():
    from pip.req import parse_requirements
    try:
        install_reqs = parse_requirements('requirements.txt', session=False)
    except TypeError:
        install_reqs = parse_requirements('requirements.txt')
    reqs = [str(ir.req) for ir in install_reqs if ir.req]
    return reqs


def find_package_files(dirpath, package, skip_exts=None):
    paths = []
    for (path, dirs, fnames) in os.walk(join(package, dirpath)):
        for fname in fnames:
            if skip_exts and any(fname.endswith(ext) for ext in skip_exts):
                continue
            fpath = join(path, fname)
            paths.append(relpath(fpath, package))
    return paths


''' Versioning:
1. Write each version to VERSION.txt
2. If the changes are significant, tag the release and push the new tag:
   $ python setup.py tag '''
def write_version_py(package_name, kwargs=None):
    version_txt = 'VERSION.txt'
    if not isfile(version_txt):
        return None

    with open(version_txt) as f:
        v = f.read().strip().split('\n')[0]

    try:
        import subprocess
        git_revision = subprocess.check_output(['git', 'rev-parse', '--short', 'HEAD']).rstrip()
    except:
        git_revision = ''
        pass
    
    try:
        git_revision = git_revision.decode('ascii')
    except:
        pass

    version_py = os.path.join(package_name, 'version.py')
    with open(version_py, 'w') as f:
        f.write((
            '# Do not edit this file, pipeline versioning is governed by git tags\n' +
            '__version__ = \'' + v + '\'\n' +
            '__git_revision__ = \'' + str(git_revision) + '\'') + '\n')
        if kwargs:
            for k, v in kwargs.items():
                f.write(k + ' = \'' + v + '\'\n')
    return v


def run_cmdl(_cmd):
    print('$ ' + _cmd)
    os.system(_cmd)


def compile_tool(tool_name, dirpath, requirements):
    if not all(isfile(join(dirpath, req)) for req in requirements):
        print('Compiling ' + tool_name)
        run_cmdl('make -C ' + dirpath)
        if not all(isfile(join(dirpath, req)) for req in requirements):
            err('Failed to compile ' + tool_name + ' (' + dirpath + ')\n')
            return False
    return True


def is_cleaning():
    return len(sys.argv) == 2 and sys.argv[1] == 'clean'



import sys,os
import gzip

class gzip_opener:
    '''
    A Python 2.6 class to handle 'with' opening of text files that may
    or may not be gzip compressed.
    '''
    def __init__(self,fname):
        self.fname = fname
    def open(self):
        return self.__enter__()
    def __enter__(self):
        if self.fname == '-':
            self.f = sys.stdin
        elif self.fname[-3:] == '.gz':
            self.f = gzip.open(os.path.expanduser(self.fname))
        else:
            self.f = open(os.path.expanduser(self.fname))
        return self.f
    def __exit__(self, type, value, traceback):
        if self.f != sys.stdin:
            self.f.close()
        return False

def filenames_to_uniq(names,new_delim='.'):
    '''
    Given a set of file names, produce a list of names consisting of the
    uniq parts of the names. This works from the end of the name.  Chunks of
    the name are split on '.' and '-'.
    
    For example:
        A.foo.bar.txt
        B.foo.bar.txt
        returns: ['A','B']
    
        AA.BB.foo.txt
        CC.foo.txt
        returns: ['AA.BB','CC']
    
    '''
    name_words = []
    maxlen = 0
    for name in names:
        name_words.append(name.replace('.',' ').replace('-',' ').strip().split())
        name_words[-1].reverse()
        if len(name_words[-1]) > maxlen:
            maxlen = len(name_words[-1])

    common = [False,] * maxlen
    for i in range(maxlen):
        last = None
        same = True
        for nameword in name_words:
            if i >= len(nameword):
                same = False
                break
            if not last:
                last = nameword[i]
            elif nameword[i] != last:
                same = False
                break
        common[i] = same

    newnames = []
    for nameword in name_words:
        nn = []
        for (i, val) in enumerate(common):
            if not val and i < len(nameword):
                nn.append(nameword[i])
        nn.reverse()
        newnames.append(new_delim.join(nn))
        
    return newnames



#!/usr/bin/env python
'''
Re-orders columns in a tab delimited file

(Reorders based upon column index, not name)
'''

import sys,os

from .support import gzip_opener

def tab_reorder(fname, column_order, delim='\t'):
    f = gzip_opener(fname).open()
    for line in f:
        if line[0] == '#':
            sys.stdout.write(line)
            continue

        cols = line.rstrip('\n').split(delim)

        cols_used = set()
        outcols = []

        for col_idx in column_order:
            if col_idx == '*':
                for i, val in enumerate(cols):
                    if not i in cols_used:
                        outcols.append(val)
                break
            else:
                outcols.append(cols[col_idx])
                if col_idx < 0:
                    cols_used.add(len(cols) + col_idx)
                else:
                    cols_used.add(col_idx)


        sys.stdout.write('%s\n' % '\t'.join(outcols))
    f.close() 
    
def usage(msg=""):
    if msg:
        print(msg)
    print(__doc__)
    print("""Usage: %s {opts} filename.tab col1,col2,col3

Columns should be specified by index:
    1,5,3 (etc)

Alternatively, ranges may be specified using from:to notation:
    1:3

Also, negative indexes may be used to indicate columns from the right side
    -3:-1

Finally, the catch-all '*' can be used to output all columns which haven't been included yet.

Note: columns start at 0

Options:
    -d delim    Use this (opposed to a tab) for the delimiter

""" % os.path.basename(sys.argv[0]))
    sys.exit(1)
    
def main(argv):
    fname = None
    delim = '\t'
    last = None

    column_order = []

    for arg in argv:
        if arg in ['-h','--help']:
            usage()
        elif last == '-d':
            delim = arg
            last = None
        elif arg in ['-d']:
            last = arg
        elif not fname and (os.path.exists(arg) or arg == '-'):
            fname = arg
        else:
            for val in arg.split(','):
                if ':' in val:
                    l,r = [int(x) for x in val.split(':')]
                    for i in range(l,r+1):
                        column_order.append(i)
                elif val != '*':
                    column_order.append(int(val))
                else:
                    column_order.append(val)
    
    if not fname or not column_order:
        usage("Missing input file or columns!")

    tab_reorder(fname, column_order, delim)
    
if __name__ == '__main__':
    main(sys.argv[1:])




"""Centralize running of external commands, providing logging and tracking.
"""
import collections
import os
import subprocess
from os.path import isfile

from ngs_utils.file_utils import file_transaction, verify_file
from ngs_utils.logger import info, err


def run(cmd, output_fpath=None, input_fpath=None, checks=None, stdout_to_outputfile=True,
        stdout_tx=True, reuse=False, stderr_fpath=None, env_vars=None):
    """Run the provided command, logging details and checking for errors.
    """
    if output_fpath and reuse:
        if verify_file(output_fpath, silent=True):
            info(output_fpath + ' exists, reusing')
            return output_fpath
        if not output_fpath.endswith('.gz') and verify_file(output_fpath + '.gz', silent=True):
            info(output_fpath + '.gz exists, reusing')
            return output_fpath

    env = os.environ.copy()
    if env_vars:
        for k, v in env_vars.items():
            if v is None:
                if k in env:
                    del env[k]
            else:
                env[k] = v

    if checks is None:
        checks = [file_nonempty_check]

    def _try_run(_cmd, _output_fpath, _input_fpath, _stderr_fpath):
        try:
            info(' '.join(str(x) for x in _cmd) if not isinstance(_cmd, str) else _cmd)
            _do_run(_cmd, checks, env, _output_fpath, _input_fpath, _stderr_fpath)
        except:
            raise

    if output_fpath:
        if isfile(output_fpath):
            os.remove(output_fpath)
    if output_fpath:
        if stdout_tx:
            with file_transaction(None, output_fpath) as tx_out_file:
                if stdout_to_outputfile:
                    cmd += ' > ' + tx_out_file
                else:
                    cmd += '\n'
                    cmd = cmd.replace(' ' + output_fpath + ' ', ' ' + tx_out_file + ' ') \
                             .replace(' "' + output_fpath + '" ', ' ' + tx_out_file + '" ') \
                             .replace(' \'' + output_fpath + '\' ', ' ' + tx_out_file + '\' ') \
                             .replace(' ' + output_fpath + '\n', ' ' + tx_out_file) \
                             .replace(' "' + output_fpath + '"\n', ' ' + tx_out_file + '"') \
                             .replace(' \'' + output_fpath + '\'\n', ' ' + tx_out_file + '\'') \
                             .replace('\n', '')
                _try_run(cmd, tx_out_file, input_fpath, stderr_fpath)
        else:
            _try_run(cmd, output_fpath, input_fpath, stderr_fpath)

    else:
        _try_run(cmd, None, input_fpath, stderr_fpath)


def find_bash():
    for test_bash in [find_cmd("bash"), "/bin/bash", "/usr/bin/bash", "/usr/local/bin/bash"]:
        if test_bash and os.path.exists(test_bash):
            return test_bash
    raise IOError("Could not find bash in any standard location. Needed for unix pipes")


def find_cmd(cmd):
    try:
        return subprocess.check_output(["which", cmd]).strip()
    except subprocess.CalledProcessError:
        return None


def _normalize_cmd_args(cmd):
    """Normalize subprocess arguments to handle list commands, string and pipes.
    Piped commands set pipefail and require use of bash to help with debugging
    intermediate errors.
    """
    if isinstance(cmd, str):
        # check for standard or anonymous named pipes
        if cmd.find(" | ") > 0 or cmd.find(">(") or cmd.find("<("):
            return "set -o pipefail; " + cmd, True, find_bash()
        else:
            return cmd, True, None
    else:
        return [str(x) for x in cmd], False, None


def _do_run(cmd, checks, env=None, output_fpath=None, input_fpath=None, _stderr_fpath=None):
    """Perform running and check results, raising errors for issues.
    """
    cmd, shell_arg, executable_arg = _normalize_cmd_args(cmd)
    s = subprocess.Popen(cmd, shell=shell_arg, executable=executable_arg,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.STDOUT, close_fds=True, env=env, encoding='utf-8')
    debug_stdout = collections.deque(maxlen=100)
    while 1:
        line = s.stdout.readline()
        if line:
            debug_stdout.append(line)
            info('  ' + line.rstrip())
        exitcode = s.poll()
        if exitcode is not None:
            for line in s.stdout:
                debug_stdout.append(line)
            if exitcode is not None and exitcode != 0:
                error_msg = " ".join(cmd) if not isinstance(cmd, str) else cmd
                error_msg += "\n"
                error_msg += "".join(debug_stdout)
                s.communicate()
                s.stdout.close()
                raise subprocess.CalledProcessError(exitcode, error_msg)
            else:
                break
    s.communicate()
    s.stdout.close()
    # Check for problems not identified by shell return codes
    if checks:
        for check in checks:
            if not check(output_fpath, input_fpath):
                raise IOError("External command failed")
        # except subprocess.CalledProcessError as e:
        #     e.returncode
        #     e.cmd


def file_nonempty_check(output_fpath=None, input_fpath=None):
    if output_fpath is None:
        return True
    ok = file_exists_check(output_fpath)
    if not ok:
        err("Did not find non-empty output file {0}".format(output_fpath))
    return ok


def file_exists_check(output_fpath=None, input_fpath=None):
    if output_fpath is None:
        return True
    ok = os.path.exists(output_fpath)
    if not ok:
        err("Did not find output file {0}".format(output_fpath))
    return ok


def file_reasonable_size(output_fpath, input_fpath):
    ok = file_exists_check(output_fpath)
    if not ok:
        return ok
    # named pipes -- we can't calculate size
    if input_fpath.strip().startswith("<("):
        return True
    if input_fpath.endswith((".bam", ".gz")):
        scale = 7.0
    else:
        scale = 10.0
    orig_size = os.path.getsize(input_fpath) / pow(1024.0, 3)
    out_size = os.path.getsize(output_fpath) / pow(1024.0, 3)
    if out_size < (orig_size / scale):
        err("Output file unexpectedly small. %.1fGb for output versus "
            "%.1fGb for the input file. This often indicates a truncated "
            "BAM file or memory errors during the run." % (out_size, orig_size))
        return False
    else:
        return True



#!/usr/bin/env python

import os
import sys
import base64
import mimetypes
import traceback
from bs4 import BeautifulSoup


def execute(html_fpath):
    if not os.path.isfile(html_fpath):
        return None
    b64_html_fpath = os.path.splitext(html_fpath)[0] + '.b64html'

    data = _html_to_dataurl(html_fpath)

    with open(b64_html_fpath, 'wt') as fp:
        fp.write(data)
    print('Saved to ' + b64_html_fpath)


def _html_to_dataurl(html_fpath):
    comp_html_unicode = _compact_html(html_fpath)
    return _convert_text(comp_html_unicode.encode('utf-8'), mime='text/html')


def _compact_html(html_fpath):
    """ 1. Embed images in base64 format
        2. Replace anchors with javascript
    """
    with open(html_fpath) as f:
        t = f.read()
        try:
            html_unicode = unicode(t, 'utf-8')
        except TypeError:
            html_unicode = unicode(t)
        except NameError:  # py3
            html_unicode = t
    base_dir = os.path.split(html_fpath)[0]
    soup = BeautifulSoup(html_unicode, 'lxml')
    for img in soup.findAll('img'):
        img_src = img['src']
        durl_img = _convert_file_contents(os.path.join(base_dir, img['src']))
        img['src'] = durl_img

    js = "javascript: void(0); document.getElementById('%s').scrollIntoView(true);"
    for anchor in soup.findAll('a'):
        if 'href' not in anchor:
            continue
        if anchor['href'].startswith('#'):
            anchor['href'] = js % anchor['href'][1:]
        else:
            del anchor['href']
    return soup.prettify()


def _convert_file_contents(fpath, mime=None):
    print('Encoding file ' + fpath)
    if not os.path.isfile(fpath):
        sys.stderr.write('File ' + fpath + ' does not exists\n')
        return ''
    if not mime:
        mimetypes.init()
        mime, enc = mimetypes.guess_type(os.path.join('file://', fpath))
        if mime is None:
            raise Exception('rfc2397: failed to determine file type')
        print('Mime: ' + str(mime) + ', encoding name: ' + str(enc))

    with open(fpath, 'rb') as fp:
        data = fp.read()
    b64_string = base64.b64encode(data)
    return 'data:%s;base64,%s' % (mime, b64_string)


def _convert_text(text, mime):
    print('Encoding ' + 'Mime: ' + str(mime))
    b64_string = base64.b64encode(text)
    return 'data:%s;base64,%s' % (mime, b64_string)


if __name__ == "__main__":
    if len(sys.argv) < 2:
        sys.stderr.write('Please, provide input HTML file. Usage: ' + __file__ + ' file.html\n')
        sys.exit(1)
    execute(sys.argv[1])





#!/usr/bin/env python

import copy
import os
import shutil
import subprocess
import sys
import tempfile
from optparse import OptionParser
from os.path import abspath, join, splitext
from os.path import exists, basename

from source.config import Config
from source.file_utils import add_suffix, adjust_path
from source.logger import critical, info
from source.prepare_args_and_cnf import determine_sys_cnf
from source.targetcov.Region import SortableByChrom
from source.utils import get_chr_lengths_from_seq

from Utils.tools import annotate_bed

""" Input: Any BED file
    Output:
        BED file with regions from input and with exactly 4 columns (or exactly 8 columns for BEDs with primers info).
        The 4th column ("gene symbol") is generated with annotate_bed.py script. If no annotation is found
        gene symbol is set to "not_a_gene_%d". If more than one annotation is found for region only one is remained
        (priory from highest to lowest: key gene, approved gene, first gene).
        Output BED is sorted using by chromosome name -> start -> end. Run standardize_bed.py --help for details about
        options.

    Usage: python standardize_bed.py [options] Input_BED_file -o Standardized_BED_file
"""


def _read_args(args_list):
    options = [
        # (['-k', '--key-genes'], dict(
        #     dest='key_genes_fpath',
        #     help='list of key genes (they are at top priority when choosing one of multiple annotations)',
        #     default='/ngs/reference_data/genomes/Hsapiens/common/az_key_genes.300.txt')
        #  ),
        # (['-a', '--approved-genes'], dict(
        #     dest='approved_genes_fpath',
        #     help='list of HGNC approved genes (they are preferable when choosing one of multiple annotations)',
        #     default='/ngs/reference_data/genomes/Hsapiens/common/HGNC_gene_synonyms.txt')
        #  ),
        # (['-e', '--ensembl-bed'], dict(
        #     dest='ensembl_bed_fpath',
        #     help='reference BED file for annotation (Ensembl)')
        #  ),
        # (['-r', '--refseq-bed'], dict(
        #     dest='refseq_bed_fpath',
        #     help='reference BED file for annotation (RefSeq)')
        #  ),
        # (['-b', '--bedtools'], dict(
        #     dest='bedtools',
        #     help='path to bedtools',
        #     default='bedtools')
        #  ),
        (['-o', '--output-bed'], dict(
            dest='output_fpath')
         ),
        (['--debug'], dict(
            dest='debug',
            help='run in a debug more (verbose output, keeping of temporary files)',
            default=False,
            action='store_true')
         ),
        (['--output-hg'], dict(
            dest='output_hg',
            help='output chromosome names in hg-style (chrM, chr1, .., chr22, chrX, chrY)',
            default=False,
            action='store_true')
         ),
        (['--output-grch'], dict(
            dest='output_grch',
            help='output chromosome names in GRCh-style (1, .., 22, X, Y, MT)',
            default=False,
            action='store_true')
         ),
        (['-g', '--genome'], dict(
            dest='genome',
            default='hg19')
         ),
    ]

    parser = OptionParser(usage='usage: %prog [options] Input_BED_file -o Standardized_BED_file',
                          description='Scripts outputs a standardized version of input BED file. '
                                      'Standardized BED: 1) has 4 or 8 fields (for BEDs with primer info);'
                                      ' 2) has HGNC approved symbol in forth column if annotation is '
                                      'possible and not_a_gene_X otherwise;'
                                      ' 3) is sorted based on chromosome name -> start -> end;'
                                      ' 4) has no duplicated regions (regions with the same chromosome, start and end), '
                                      'the only exception is _CONTROL_ regions.')
    for args, kwargs in options:
        parser.add_option(*args, **kwargs)
    (opts, args) = parser.parse_args(args_list)

    if len(args) != 1:
        parser.print_help(file=sys.stderr)
        sys.exit(1)

    cnf = Config(opts.__dict__, determine_sys_cnf(opts), {})

    work_dirpath = tempfile.mkdtemp()
    info('Creating a temporary working directory ' + work_dirpath)
    if not exists(work_dirpath):
        os.mkdir(work_dirpath)

    input_bed_fpath = abspath(args[0])
    info('Input: ' + input_bed_fpath)

    output_bed_fpath = adjust_path(cnf.output_fpath)
    info('Writing to: ' + output_bed_fpath)

    # process configuration
    # for k, v in opts.__dict__.items():
    #     if k.endswith('fpath') and verify_file(v, is_critical=True):
    #         opts.__dict__[k] = verify_file(v, k)
    if cnf.output_grch and cnf.output_hg:
        info('you cannot specify --output-hg and --output-grch simultaneously!')
    # if not which(opts.bedtools):
    #     info('bedtools executable not found, please specify correct path (current is %s)! '
    #         'Did you forget to execute "module load bedtools"?' % opts.bedtools)

    # if opts.debug:
    #     info('Configuration: ')
    #     for k, v in opts.__dict__.items():
    #         info('\t' + k + ': ' + str(v))
    info()

    # opts.ensembl_bed_fpath = verify_file(opts.ensembl_bed_fpath or \
    #     ('/ngs/reference_data/genomes/Hsapiens/' + opts.genome + '/bed/Exons/Exons.with_genes.bed'))

    # opts.refseq_bed_fpath = verify_file(opts.refseq_bed_fpath or \
    #     ('/ngs/reference_data/genomes/Hsapiens/' + opts.genome + '/bed/Exons/RefSeq/RefSeq_CDS_miRNA.all_features.bed'))

    return input_bed_fpath, output_bed_fpath, work_dirpath, cnf


class BedParams:
    GRCh_to_hg = {'MT': 'chrM', 'X': 'chrX', 'Y': 'chrY'}
    for i in range(1, 23):
        GRCh_to_hg[str(i)] = 'chr' + str(i)
    hg_to_GRCh = dict((v, k) for k, v in GRCh_to_hg.items())

    def __init__(self, header=list(), controls=list(), GRCh_names=None, n_cols_needed=None):
        self.header = header
        self.controls = controls
        self.GRCh_names = GRCh_names
        self.n_cols_needed = n_cols_needed

    @staticmethod
    def calc_n_cols_needed(line):
        columns = line.strip().split('\t')
        if len(columns) < 8:
            return 4
        elif columns[6].isdigit() and columns[7].isdigit():  # primers info
            return 8
        else:
            return 4


class Region(SortableByChrom):
    GRCh_names = None
    n_cols_needed = 4
    approved_genes = []
    key_genes = []

    def __init__(self, chrom, chrom_ref_order, start, end):
        SortableByChrom.__init__(self, chrom, chrom_ref_order)
        self.start = start
        self.end = end
        self.symbol = None
        self.type = None
        self.rest = None

    def set_symbol(self, symbol):
        self.symbol = symbol
        if self.symbol in self.key_genes:
            self.type = 'key'
        elif self.symbol in self.approved_genes:
            self.type = 'approved'
        else:
            self.type = 'not_approved'

    def __str__(self):
        fs = [BedParams.hg_to_GRCh[self.chrom] if self.GRCh_names else self.chrom,
              str(self.start), str(self.end), self.symbol] + (self.rest if self.n_cols_needed > 4 else [])
        return '\t'.join(fs)

    def get_key(self):
        return SortableByChrom.get_key(self), self.start, self.end, self.is_control(), self.symbol

    def is_control(self):
        if self.symbol and self.symbol.startswith('_CONTROL'):
            return True
        return False

    def __lt__(self, other):
        return self.get_key() < other.get_key()

    def __eq__(self, other):
        return self.get_key() == other.get_key()

    def __hash__(self):
        return hash(self.get_key())


def _preprocess(cnf, bed_fpath, work_dirpath, chrom_order):
    bed_params = BedParams()
    output_fpath = __intermediate_fname(work_dirpath, bed_fpath, 'prep')
    info('preprocessing: ' + bed_fpath + ' --> ' + output_fpath)
    with open(bed_fpath, 'r') as in_f:
        with open(output_fpath, 'w') as out_f:
            for line in in_f:
                if line.startswith('#') or line.startswith('track') or line.startswith('browser'):  # header
                    bed_params.header.append(line if line.startswith('#') else '#' + line)
                else:
                    cur_ncn = BedParams.calc_n_cols_needed(line)
                    if bed_params.n_cols_needed is not None and cur_ncn != bed_params.n_cols_needed:
                        critical('number and type of columns should be the same on all lines!')
                    bed_params.n_cols_needed = cur_ncn
                    if line.startswith('chr'):
                        if bed_params.GRCh_names is not None and bed_params.GRCh_names:
                            critical('mixing of GRCh and hg chromosome names!')
                        bed_params.GRCh_names = False
                        if line.startswith('chrMT'):  # common misprint, correcting chrMT --> chrM
                            processed_line = '\t'.join(['chrM'] + line.split('\t')[1:])
                        else:
                            processed_line = line
                    elif line.split('\t')[0] in BedParams.GRCh_to_hg:  # GRCh chr names
                        if bed_params.GRCh_names is not None and not bed_params.GRCh_names:
                            critical('mixing of GRCh and hg chromosome names!')
                        bed_params.GRCh_names = True
                        processed_line = '\t'.join([BedParams.GRCh_to_hg[line.split('\t')[0]]] + line.split('\t')[1:])
                    else:
                        critical('incorrect chromosome name!')

                    entries = processed_line.strip().split('\t')
                    chrom = entries[0]
                    start = int(entries[1])
                    end = int(entries[2])
                    r = Region(chrom, chrom_order.get(chrom), start, end)
                    if r.is_control():
                        r.set_symbol(entries[3] if len(entries) > 3 else '{0}:{1}-{2}'.format(chrom, start, end))
                        r.rest = entries[4:] if len(entries) > 4 else None
                        bed_params.controls.append(r)
                    else:
                        out_f.write(processed_line)
    return output_fpath, bed_params


def _annotate(bed_fpath, work_dirpath, cnf):
    annotated_files = []
    input_fpath = bed_fpath
    references = [('RefSeq', cnf.genome.features), ('Ensembl', cnf.genome.ensembl)]

    for id, (db_name, db_bed_fpath) in enumerate(references):
        output_fpath = __intermediate_fname(work_dirpath, bed_fpath, 'ann_' + db_name.lower())
        info('annotating based on {db_name}: {bed_fpath} --> {output_fpath}'.format(**locals()))
        annotate_bed_py = sys.executable + ' ' + splitext(annotate_bed.__file__)[0] + '.py'

        cmdline = '{annotate_bed_py} {input_fpath} --reference {db_bed_fpath} -o {output_fpath} --genome {cnf.genome}'.format(**locals())
        __call(cnf, cmdline)

        if id < len(references) - 1:
            if cnf.debug:
                info("filtering annotated and not annotated regions into separate files:")
            only_annotated_bed = __intermediate_fname(work_dirpath, bed_fpath, 'only_ann_' + db_name.lower())
            not_annotated_bed = __intermediate_fname(work_dirpath, bed_fpath, 'not_ann_' + db_name.lower())
            with open(only_annotated_bed, 'w') as out:
                cmdline = 'grep -v -E "\.$" {output_fpath}'.format(**locals())
                if cnf.debug:
                    info(cmdline + ' > ' + only_annotated_bed)
                subprocess.call(cmdline, shell=True, stdout=out)
            with open(not_annotated_bed, 'w') as out:
                cmdline = 'grep -E "\.$" {output_fpath}'.format(**locals())
                if cnf.debug:
                    info(cmdline + ' > ' + not_annotated_bed)
                subprocess.call(cmdline, shell=True, stdout=out)
            if not cnf.debug:
                os.remove(output_fpath)
            output_fpath = only_annotated_bed
            input_fpath = not_annotated_bed
        annotated_files.append(output_fpath)
        if id != 0 and not cnf.debug:
            os.remove(input_fpath)

    return annotated_files


def _postprocess(input_fpath, annotated_fpaths, bed_params, output_bed_fpath, cnf, chrom_order):
    '''
    1. Sorts.
    1. Chooses appropriate number of columns (4 or 8 for BEDs with primers).
    2. Removes duplicates.
    '''
    info('postprocessing (sorting, cutting, removing duplicates)')

    key_genes = []
    with open(adjust_path(cnf.key_genes), 'r') as f:
        for line in f:
            key_genes.append(line.strip())
    approved_genes = []
    if cnf.hgnc:
        with open(adjust_path(cnf.hgnc), 'r') as f:
            f.readline()  # header
            for line in f:
                approved_genes.append(line.split('\t')[0])

    Region.GRCh_names = bed_params.GRCh_names
    if cnf.output_grch:
        Region.GRCh_names = True
        if cnf.debug and not bed_params.GRCh_names:
            info('Changing chromosome names from hg-style to GRCh-style.')
    if cnf.output_hg:
        Region.GRCh_names = False
        if cnf.debug and bed_params.GRCh_names:
            info('Changing chromosome names from GRCh-style to hg-style.')
    Region.n_cols_needed = bed_params.n_cols_needed
    Region.key_genes = key_genes
    Region.approved_genes = approved_genes

    input_regions = set()  # we want only unique regions
    with open(adjust_path(input_fpath)) as f:
        for line in f:
            entries = line.strip().split('\t')
            chrom = entries[0]
            start = int(entries[1])
            end = int(entries[2])
            r = Region(chrom, chrom_order.get(chrom), start, end)
            r.set_symbol(entries[3] if len(entries) > 3 else '{0}:{1}-{2}'.format(chrom, start, end))
            r.rest = entries[4:] if len(entries) > 4 else None
            input_regions.add(r)

    annotated_regions = []
    for annotated_fpath in annotated_fpaths:
        with open(adjust_path(annotated_fpath)) as f:
            for line in f:
                entries = line.strip().split('\t')
                chrom = entries[0]
                start = int(entries[1])
                end = int(entries[2])
                r = Region(chrom, chrom_order.get(chrom), start, end)
                r.set_symbol(entries[3] if len(entries) > 3 else '{0}:{1}-{2}'.format(chrom, start, end))
                r.rest = entries[4:] if len(entries) > 4 else None
                annotated_regions.append(r)

    # starting to output result
    with open(adjust_path(output_bed_fpath), 'w') as f:
        for line in bed_params.header:
            f.write(line)

        annotated_regions.sort()
        i = 0
        prev_region = None
        not_a_gene_count = 0
        solid_regions = []
        prev_is_solid = False
        all_regions = []
        for cur_region in sorted(list(input_regions) + bed_params.controls):
            if not cur_region.is_control():
                assert annotated_regions[i] == cur_region, str(cur_region) + ' != ' + str(annotated_regions[i]) + '(i=%d)' % i
                if annotated_regions[i].symbol != '.':
                    cur_region.set_symbol(annotated_regions[i].symbol)
                else:
                    if prev_region is None or \
                       prev_region.chrom != cur_region.chrom or not prev_region.symbol.startswith("not_a_gene"):
                        not_a_gene_count += 1
                    cur_region.set_symbol("not_a_gene_%d" % not_a_gene_count)
                i += 1
                ambiguous_regions = [cur_region]
                while i < len(annotated_regions) and annotated_regions[i] == cur_region:  # processing duplicates
                    if annotated_regions[i].symbol != '.' and annotated_regions[i].symbol != cur_region.symbol:
                        duplicate = copy.deepcopy(cur_region)
                        duplicate.set_symbol(annotated_regions[i].symbol)
                        if duplicate.type == 'approved' and cur_region.type == 'not_approved':
                            cur_region = duplicate
                            ambiguous_regions = [cur_region]
                        elif annotated_regions[i].type == 'key' and cur_region.type != 'key':
                            cur_region = duplicate
                            ambiguous_regions = [cur_region]
                            if cnf.debug:
                                info('key gene priority over approved gene was used')
                        elif annotated_regions[i].type == cur_region.type:
                            ambiguous_regions.append(duplicate)
                    i += 1
                if len(ambiguous_regions) == 1:
                    if not prev_is_solid:
                        solid_regions.append(cur_region)
                    prev_is_solid = True
                    all_regions.append(cur_region)
                else:
                    if prev_is_solid:
                        solid_regions.append(prev_region)
                    prev_is_solid = False
                    all_regions.append(ambiguous_regions)
            else:
                all_regions.append(cur_region)
            prev_region = cur_region

        # outputting results
        cur_solid_id = -1
        for entry in all_regions:
            if isinstance(entry, list):  # list of ambiguous regions
                cur_region = entry[0]
                while cur_solid_id + 1 < len(solid_regions) and cur_region > solid_regions[cur_solid_id + 1]:
                    cur_solid_id += 1
                found = False
                if cur_solid_id >= 0 and cur_region > solid_regions[cur_solid_id] \
                        and cur_region.chrom == solid_regions[cur_solid_id].chrom:
                    prev_solid = solid_regions[cur_solid_id]
                    for cur_region in entry:
                        if cur_region.symbol == prev_solid.symbol:
                            found = True
                            if cnf.debug:
                                info('gene name was chosen based on previous solid region')
                            break
                if not found and cur_solid_id + 1 < len(solid_regions) and cur_region < solid_regions[cur_solid_id + 1] \
                        and cur_region.chrom == solid_regions[cur_solid_id + 1].chrom:
                    next_solid = solid_regions[cur_solid_id + 1]
                    for cur_region in entry:
                        if cur_region.symbol == next_solid.symbol:
                            found = True
                            if cnf.debug:
                                info('gene name was chosen based on next solid region')
                            break
                if not found:
                    cur_region = entry[0]
            else:
                cur_region = entry
            f.write(str(cur_region) + '\n')  # automatically outputs correct number of columns and GRCh/hg names


def __intermediate_fname(work_dir, fname, suf):
    output_fname = add_suffix(fname, suf)
    return join(work_dir, basename(output_fname))


def __call(cnf, cmdline, output_fpath=None):
    stdout = open(output_fpath, 'w') if output_fpath else None
    stderr = None if cnf.debug else open('/dev/null', 'w')
    if cnf.debug:
        info(cmdline)
    ret_code = subprocess.call(cmdline, shell=True, stdout=stdout, stderr=stderr, stdin=None)
    return ret_code


def main():
    input_bed_fpath, output_bed_fpath, work_dirpath, cnf = _read_args(sys.argv[1:])

    chr_lengths = get_chr_lengths_from_seq(cnf.genome.seq)
    chrom_order = {c: i for i, (c, l) in enumerate(chr_lengths)}

    preprocessed_fpath, bed_params = _preprocess(cnf, input_bed_fpath, work_dirpath, chrom_order)
    annotated_fpaths = _annotate(preprocessed_fpath, work_dirpath, cnf)

    _postprocess(preprocessed_fpath, annotated_fpaths, bed_params, output_bed_fpath, cnf, chrom_order)
    if not cnf.debug:
        for f in [preprocessed_fpath] + annotated_fpaths:
            os.remove(f)

    try:
        shutil.rmtree(work_dirpath)
    except OSError:
        pass

if __name__ == '__main__':
    main()



#!/usr/bin/env python

from os.path import abspath, dirname, realpath, join
import sys
from collections import OrderedDict, defaultdict

from ngs_utils.bed_utils import count_bed_cols


def main():
    if len(sys.argv) < 2:
        sys.exit('Usage: ' + __file__ + ' bed_file > merged_bed_file')

    summarize_by_genes = True  # len(sys.argv) > 2
    # sys.stderr.write('Setting summarize_by_genes to ' + str(summarize_by_genes) + '\n')

    num_bed_cols = count_bed_cols(sys.argv[1])
    if num_bed_cols < 3:
        sys.exit('Incorrect number of fields: ' + str(num_bed_cols) + '. Should be at least 3.')
    if num_bed_cols < 7:
        summarize_by_genes = False
        sys.stderr.write('less than 7 columns in BED; no summarizing by genes\n')

    gene_by_chrom_and_name = OrderedDict()

    total_lines = 0
    feature_counter = defaultdict(int)
    with open(sys.argv[1]) as inp:
        for l in inp:
            if not l:
                pass
            elif l.startswith('#') or '\t' not in l:
                sys.stdout.write(l)
            else:
                fields = l.replace('\n', '').split('\t')
                if len(fields) != num_bed_cols:
                    sys.stderr.write('Error: number of fields inconsistent. Expected ' + str(num_bed_cols) + ', got ' + str(len(fields)) + ' at + ' + ' | '.join(fields) + '\n')
                    sys.exit(1)
                else:
                    chrom, start, end = fields[:3]
                    start, end = int(start), int(end)
                    gname = fields[3] if num_bed_cols >= 4 else '.'
                    strand = fields[5] if num_bed_cols >= 6 else None
                    (feature, biotype) = fields[6:8] if num_bed_cols >= 8 else (None, None)
                    if feature:
                        feature_counter[feature] += 1

                    gene = gene_by_chrom_and_name.get((chrom, gname, strand))
                    if gene is None:
                        gene = Gene(gname, chrom, strand, 'Gene', biotype)
                        gene_by_chrom_and_name[(chrom, gname, strand)] = gene

                    if feature in ['Gene', 'Multi_Gene']:  # in fact '*Gene' features in BED files are optional
                        if gene.already_met_gene_feature_for_this_gene:
                            # sys.stderr.write(gene.name + ' is duplicating: ' + str(gene) + '\n')
                            # sys.exit(1)
                            # miltiple records for gene, picking the lowest start and the biggest end
                            gene.start = min(gene.start, start)
                            gene.end = max(gene.end, end)
                            gene.biotype = merge_fields(gene.biotype, biotype)
                            # assert gene.strand == strand, 'Prev gene strand is ' + gene.strand + ', new strand is ' + strand + ' gene is ' + gene.name

                        assert gene.strand == strand, str(gene) + ' strand is not ' + strand
                        gene.feature = feature
                        gene.start = start
                        gene.end = end
                        gene.biotype = biotype
                        gene.already_met_gene_feature_for_this_gene = True

                    elif feature in [None, '.', 'CDS', 'Exon', 'UTR/Intron/Decay']:
                        assert gene.strand == strand, str(gene) + ' strand is not ' + strand
                        gene.regions.append(Exon(int(start), int(end), biotype, feature))

            total_lines += 1
            if total_lines % 10000 == 0:
                sys.stderr.write('processed ' + str(total_lines // 1000) + 'k lines\n')
                sys.stderr.flush()

    sys.stderr.write('Processed ' + str(total_lines) + ' lines, found ' + str(len(gene_by_chrom_and_name)) + ' unique genes.\n')
    if feature_counter:
        sys.stderr.write('Features:\n')
        for ft, cnt in feature_counter.items():
            sys.stderr.write('  ' + ft + ': ' + str(cnt) + '\n')
    sys.stderr.write('\n')

    genes = []
    for gene in gene_by_chrom_and_name.values():
        if gene.sort_regions() is not None:
            genes.append(gene)

    sys.stderr.write('Merging regions...\n')
    final_regions = []
    for gene in sorted(genes, key=lambda g: g.get_key()):
        if summarize_by_genes and gene.name != '.':
            final_regions.append((gene.chrom, gene.start, gene.end, gene.name, gene.strand, gene.feature, gene.biotype))

        merged_regions = gene.merge_regions()
        for r in merged_regions:
            final_regions.append((gene.chrom, r.start, r.end, gene.name, gene.strand, r.feature, r.biotype))

    sys.stderr.write('Merged, regions after merge: ' + str(len(final_regions)) + ', saving...\n')

    for chrom, start, end, gname, strand, feature, biotype in final_regions:
        fs = [chrom, str(start), str(end), gname, '.', strand or '.', feature or '.', biotype or '.']
        sys.stdout.write('\t'.join(fs[:num_bed_cols]) + '\n')
    sys.stderr.write('Saved\n')


class Exon:
    def __init__(self, start, end, biotype=None, feature=None):
        self.start = start
        self.end = end
        self.biotype = biotype
        self.feature = feature

    def __repr__(self):
        return str(self.start) + '-' + str(self.end) + ',' + str(self.biotype) + ', ' + str(self.feature)


CHROMS = [('Y', 23), ('X', 24), ('M', 0)]
for i in range(22, 0, -1):
    CHROMS.append((str(i), i))


class Gene:
    def __init__(self, name, chrom, strand=None, feature=None, biotype=None):
        self.name = name
        self.chrom = chrom
        self.__chrom_key = self.__make_chrom_key()
        self.strand = strand
        self.start = None
        self.end = None
        self.feature = feature
        self.biotype = biotype
        self.already_met_gene_feature_for_this_gene = False  # some BED files can contain '*Gene' features, so we can take start and end from them

        self.regions = []

    def __make_chrom_key(self):
        chr_remainder = self.chrom
        if self.chrom.startswith('chr'):
            chr_remainder = self.chrom[3:]
        for (c, i) in CHROMS:
            if chr_remainder == c:
                return i
            elif chr_remainder.startswith(c):
                return i + 24

        sys.stderr.write('Cannot parse chromosome ' + self.chrom + '\n')
        return None

    def get_key(self):
        return self.__chrom_key, self.start, (0 if 'Gene' in self.feature else 1), self.end, self.name

    def sort_regions(self):
        self.regions = sorted(self.regions, key=lambda r: (r.start, r.end))
        if self.start is None or self.end is None:
            if not self.regions:
                return None  # no coordinates and no exons to infer coordinates
            else:
                self.start = self.regions[0].start
                self.end = self.regions[-1].end
        return self.regions

    def merge_regions(self):
        if len(self.regions) == 0:
            sys.stderr.write('Error: no sub-regions of ' + str(self) + '\n')
            sys.exit(1)

        non_overlapping_regions = [self.regions[0]]

        for r in self.regions[1:]:
            if r.start > non_overlapping_regions[-1].end:
                non_overlapping_regions.append(r)
            else:
                prev_r = non_overlapping_regions[-1]
                prev_r.end = r.end
                prev_r.biotype = merge_fields(prev_r.biotype, r.biotype)
                prev_r.feature = merge_fields(prev_r.feature, r.feature)

        self.regions = non_overlapping_regions
        return non_overlapping_regions

    def __repr__(self):
        return self.chrom + ':' + str(self.start) + '-' + str(self.end) + ',' + str(self.name)

    def __str__(self):
        return self.__repr__()


def merge_fields(consensus_field, other_field):
    if not consensus_field:
        consensus_field = other_field
    else:
        consensus_field = ','.join(set(consensus_field.split(',')) | set(other_field.split(',')))
    return consensus_field


if __name__ == '__main__':
    main()


import gffutils
import tempfile
import os
import random
import gzip
from ngs_utils.file_utils import file_exists, file_transaction
from ngs_utils.logger import debug


def guess_infer_extent(gtf_file):
    """
    guess if we need to use the gene extent option when making a gffutils
    database by making a tiny database of 1000 lines from the original
    GTF and looking for all of the features
    """
    _, ext = os.path.splitext(gtf_file)
    tmp_out = tempfile.NamedTemporaryFile(suffix=".gtf", delete=False).name
    with open(tmp_out, "w") as out_handle:
        count = 0
        in_handle = open(gtf_file) if ext != ".gz" else gzip.open(gtf_file)
        for line in in_handle:
            if count > 1000:
                break
            out_handle.write(line)
            count += 1
        in_handle.close()
    db = gffutils.create_db(tmp_out, dbfn=":memory:", infer_gene_extent=False)
    os.remove(tmp_out)
    features = [x for x in db.featuretypes()]
    if "gene" in features and "transcript" in features:
        return False
    else:
        return True

def get_gtf_db(gtf, in_memory=False):
    """
    create a gffutils DB
    """
    db_file = gtf + '.db'
    if gtf.endswith('.gz'):
        db_file = gtf[:-3] + '.db'
    if file_exists(db_file):
        return gffutils.FeatureDB(db_file)
    db_file = ':memory:' if in_memory else db_file
    if in_memory or not file_exists(db_file):
        debug('GTF database does not exist, creating...')
        infer_extent = guess_infer_extent(gtf)
        db = gffutils.create_db(gtf, dbfn=db_file,
                                infer_gene_extent=infer_extent)
        return db
    else:
        return gffutils.FeatureDB(db_file)

def gtf_to_bed(gtf, alt_out_dir=None):
    """
    create a BED file of transcript-level features with attached gene name
    or gene ids
    """
    out_file = os.path.splitext(gtf)[0] + '.bed'
    if file_exists(out_file):
        return out_file
    if not os.access(os.path.dirname(out_file), os.W_OK | os.X_OK):
        if not alt_out_dir:
            raise IOError('Cannot write transcript BED output file %s' % out_file)
        else:
            out_file = os.path.join(alt_out_dir, os.path.basename(out_file))
    with open(out_file, "w") as out_handle:
        db = get_gtf_db(gtf)
        for feature in db.features_of_type('transcript', order_by=("seqid", "start", "end")):
            chrom = feature.chrom
            start = feature.start
            end = feature.end
            attributes = feature.attributes.keys()
            strand = feature.strand
            name = (feature['gene_name'][0] if 'gene_name' in attributes else
                    feature['gene_id'][0])
            line = "\t".join([str(x) for x in [chrom, start, end, name, ".",
                                               strand]])
            out_handle.write(line + "\n")
    return out_file

def complete_features(db):
    """
    iterator returning features which are complete (have a 'gene_id' and a
    'transcript_id')
    """
    for feature in db.all_features():
        gene_id = feature.attributes.get('gene_id', [None])[0]
        transcript_id = feature.attributes.get('transcript_id', [None])[0]
        if gene_id and transcript_id and feature.featuretype != "transcript":
            yield feature

def partition_gtf(gtf, coding=False, out_file=False):
    """
    return a GTF file of all non-coding or coding transcripts. the GTF must be annotated
    with gene_biotype = "protein_coding" or to have the source column set to the
    biotype for all coding transcripts. set coding to
    True to get only the coding, false to get only the non-coding
    """
    if out_file and file_exists(out_file):
        return out_file
    if not out_file:
        out_file = tempfile.NamedTemporaryFile(delete=False,
                                               suffix=".gtf").name

    if coding:
        pred = lambda biotype: biotype and biotype == "protein_coding"
    else:
        pred = lambda biotype: biotype and biotype != "protein_coding"

    biotype_lookup = _biotype_lookup_fn(gtf)

    db = get_gtf_db(gtf)
    with file_transaction(out_file) as tx_out_file:
        with open(tx_out_file, "w") as out_handle:
            for feature in db.all_features():
                biotype = biotype_lookup(feature)
                if pred(biotype):
                    out_handle.write(str(feature) + "\n")
    return out_file

def get_coding_noncoding_transcript_ids(gtf):
    """
    return a set of coding and non-coding transcript_ids from a GTF
    """
    coding_gtf = partition_gtf(gtf, coding=True)
    coding_db = get_gtf_db(coding_gtf)
    coding_ids = set([x['transcript_id'][0] for x in coding_db.all_features()
                  if 'transcript_id' in x.attributes])
    noncoding_gtf = partition_gtf(gtf)
    noncoding_db = get_gtf_db(noncoding_gtf)
    noncoding_ids = set([x['transcript_id'][0] for x in noncoding_db.all_features()
                     if 'transcript_id' in x.attributes])
    return coding_ids, noncoding_ids

def get_gene_source_set(gtf):
    """
    get a dictionary of the set of all sources for a gene
    """
    gene_to_source = {}
    db = get_gtf_db(gtf)
    for feature in complete_features(db):
        gene_id = feature['gene_id'][0]
        sources = gene_to_source.get(gene_id, set([])).union(set([feature.source]))
        gene_to_source[gene_id] = sources
    return gene_to_source

def get_transcript_source_set(gtf):
    """
    get a dictionary of the set of all sources of the gene for a given
    transcript
    """
    gene_to_source = get_gene_source_set(gtf)
    transcript_to_source = {}
    db = get_gtf_db(gtf)
    for feature in complete_features(db):
        gene_id = feature['gene_id'][0]
        transcript_to_source[feature['transcript_id'][0]] = gene_to_source[gene_id]
    return transcript_to_source

def _biotype_lookup_fn(gtf):
    """
    return a function that will look up the biotype of a feature
    this checks for either gene_biotype or biotype being set or for the source
    column to have biotype information
    """
    db = get_gtf_db(gtf)
    sources = set([feature.source for feature in db.all_features()])
    gene_biotypes = set([feature.attributes.get("gene_biotype", [None])[0]
                         for feature in db.all_features()])
    biotypes = set([feature.attributes.get("biotype", [None])[0]
                    for feature in db.all_features()])
    if "protein_coding" in sources:
        return lambda feature: feature.source
    elif "protein_coding" in biotypes:
        return lambda feature: feature.attributes.get("biotype", [None])[0]
    elif "protein_coding" in gene_biotypes:
        return lambda feature: feature.attributes.get("gene_biotype", [None])[0]
    else:
        return None

def tx2genefile(gtf, out_file=None):
    """
    write out a file of transcript->gene mappings.
    use the installed tx2gene.csv if it exists, else write a new one out
    """
    installed_tx2gene = os.path.join(os.path.dirname(gtf), "tx2gene.csv")
    if file_exists(installed_tx2gene):
        return installed_tx2gene
    if file_exists(out_file):
        return out_file
    with file_transaction(out_file) as tx_out_file:
        with open(tx_out_file, "w") as out_handle:
            for k, v in transcript_to_gene(gtf).items():
                out_handle.write(",".join([k, v]) + "\n")
    return out_file

def transcript_to_gene(gtf):
    """
    return a dictionary keyed by transcript_id of the associated gene_id
    """
    gene_lookup = {}
    for feature in complete_features(get_gtf_db(gtf)):
        gene_id = feature.attributes.get('gene_id', [None])[0]
        transcript_id = feature.attributes.get('transcript_id', [None])[0]
        gene_lookup[transcript_id] = gene_id
    return gene_lookup



import contextlib
import os
import subprocess

from cluster_helper.cluster import ClusterView as CV
from joblib import Parallel, delayed

from ngs_utils.file_utils import safe_mkdir
from ngs_utils.logger import debug, err
from ngs_utils.utils import is_cluster


class ParallelCfg:
    def __init__(self,
                 scheduler=None,
                 queue=None,
                 resources=None,
                 threads=None,
                 tag=None,
                 local=False):
        self.scheduler = scheduler
        self.queue = queue
        self.threads = threads or 1
        self.extra_params = dict()
    
        self.extra_params['run_local'] = local or not is_cluster()
        if tag:
            self.extra_params['tag'] = tag
        if resources:
            self.extra_params['resources'] = resources

    def set_tag(self, tag):
        self.extra_params['tag'] = tag

    def num_jobs(self, n_samples):
        return min(self.threads, n_samples)

    def cores_per_job(self, n_jobs):
        return max(1, self.threads // n_jobs)

    def get_cluster_params(self, n_samples):
        return dict(
            scheduler=self.scheduler,
            queue=self.queue,
            num_jobs=self.num_jobs(n_samples),
            cores_per_job=self.cores_per_job(self.num_jobs(n_samples)),
            start_wait=999,
            extra_params=self.extra_params)


def get_parallel_view(n_samples, parallel_cfg):
    if parallel_cfg.scheduler and parallel_cfg.threads > 1:
        debug('Starting' + (' test' if not is_cluster() else '') + ' cluster (scheduler: ' + parallel_cfg.scheduler + ', queue: ' + parallel_cfg.queue + ') '
              'using ' + str(parallel_cfg.num_jobs(n_samples)) + ' nodes, ' + str(parallel_cfg.cores_per_job(n_samples)) + ' threads per each sample')
        return ClusterView(n_samples, parallel_cfg)
    else:
        debug('Running locally using ' + str(parallel_cfg.num_jobs(n_samples)) + ' thread(s)')
        return ThreadedView(n_samples, parallel_cfg)


@contextlib.contextmanager
def parallel_view(n_samples, parallel_cfg, work_dir):
    prev_dir = os.getcwd()
    os.chdir(safe_mkdir(work_dir))
    view = get_parallel_view(n_samples, parallel_cfg)
    os.chdir(prev_dir)
    try:
        yield view
    finally:
        view.stop()


class BaseView:
    def __init__(self, n_samples, parallel_cfg):
        self.n_samples = n_samples
        self.parallel_cfg = parallel_cfg
        self.num_jobs = parallel_cfg.num_jobs(n_samples)
        self.cores_per_job = parallel_cfg.cores_per_job(n_samples)
        self._view = None

    def run(self, fn, param_lists):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError


class ClusterView(BaseView):
    def __init__(self, n_samples, parallel_cfg):
        BaseView.__init__(self, n_samples, parallel_cfg)
        self._view = CV(**parallel_cfg.get_cluster_params(n_samples))
        debug('Starting cluster with ' + str(self.num_jobs) + ' open nodes, ' + str(self.cores_per_job) + ' cores per node')

    def run(self, fn, param_lists):
        if self.n_samples == 0:
            return []
        assert self.n_samples == len(param_lists)
        n_params = len(param_lists[0])
        for sample_i, params in enumerate(param_lists):
            if params is None:
                err('Parameter list for sample ' + str(sample_i) + ' is None')
            if len(params) != n_params:
                err('Parameter list for sample ' + str(sample_i) + ' (' + str(len(params)) +
                    ') does not equal to the one for sample 1 (' + str(n_params) + ')')
        res = self._view.view.map(fn, *([params[param_i] for params in param_lists] for param_i in range(n_params)))
        return res

    def stop(self):
        self._view.stop()


class ThreadedView(BaseView):
    def __init__(self, n_samples, parallel_cfg):
        BaseView.__init__(self, n_samples, parallel_cfg)
        self._view = Parallel(n_jobs=self.num_jobs)

    def run(self, fn, param_lists):
        debug('Starting multithreaded function' + str(fn))
        assert self.n_samples == len(param_lists)
        return self._view(delayed(fn)(*params) for params in param_lists)

    def stop(self):
        return


@contextlib.contextmanager
def with_chdir(dirpath):
    prev_dir = os.getcwd()
    try:
        os.chdir(dirpath)
        yield dirpath
    finally:
        os.chdir(prev_dir)



class BaseSample:
    def __init__(self, name=None, dirpath=None, work_dir=None, bam=None, bed=None, vcf=None, genome=None,
                 targqc_dirpath=None, clinical_report_dirpath=None,
                 normal_match=None, sv_fpath=None, sv_bed=None,
                 l_fpath=None, r_fpath=None, **kwargs):
        self.name = name
        self.dirpath = dirpath
        self.work_dir = work_dir
        self.bam = bam
        self.l_fpath = l_fpath
        self.r_fpath = r_fpath
        self.is_wgs = False
        self.vcf = vcf
        self.phenotype = None
        self.gender = None
        self.genome = None
        self.var_dirpath = None
        self.normal_match = normal_match
        self.min_af = None
        self.sv_fpath = sv_fpath
        self.targqc_dirpath = targqc_dirpath
        self.clinical_html = None
        for k, v in kwargs.items():
            self.__dict__[k] = v

    def __cmp__(self, other):
        return self.key_to_sort().__cmp__(other.key_to_sort())

    def key_to_sort(self):
        parts = []

        cur_part = []
        prev_was_num = False

        for c in self.name:
            if prev_was_num == c.isdigit() and c not in ['-', '.']:  # same type of symbol, but not - or .
                cur_part.append(c)
            else:
                if cur_part:
                    part = ''.join(cur_part)
                    if prev_was_num:
                        part = int(part)
                    parts.append(part)
                    cur_part = []

                if c in ['-', '.']:
                    pass
                else:
                    if c.isdigit():
                        prev_was_num = True
                    else:
                        prev_was_num = False
                    cur_part.append(c)
        if cur_part:
            part = ''.join(cur_part)
            if prev_was_num:
                part = int(part)
            parts.append(part)

        return tuple(parts)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name

    @classmethod
    def load(cls, data):
        sample = cls(**data)
        sample.__dict__ = data
        return sample
